""" Module responsible for data reading in DMT.

The functions here read from many different formats into a DMT-DataFrame. As reading and writing is always very close, recommended and often used save functions for DataFrames are also given.

Functions
---------
read_data(filename)
    Reads a given file into the internal DMT dataframe, the file's extension determines the exact method which is then called for reading.

read_hdf(filename, key, convert_cmplx)
    Reads in a .hdf file into the internal DMT format.

read_elpa(filename)
    Reads in a .elpa file into the internal DMT format.

read_mdm(filename)
    Reads in a .mdm file into the internal DMT format.

read_csv(filename)
    Read .csv file and generate a DMT dataframe from it.

read_feather(filename)
    Read .feather file and generate a DMT dataframe from it.

read_DEVICE_bin(filename)
    Reads DEVICE binaries. Here the internal spacial data is saved. Returns a Dataframe.

read_ADS_bin(filename)
    Reads a ADS raw data file into a DMT dataframe.

read_tikz_file(filename, col_x=None, col_y=None)
    Reads a tikz plot file into a DMT dataframe.

save_hdf(df, save_dir, filename)
    Save the dataframe df into save_dir as filename.h5 .

save_elpa(fname, ELPA, cols, firstline)
    Save data as a elpa file.

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import pandas as pd
import warnings
import numpy as np
import re
import _pickle as pickle
import struct
from pathlib import Path
from DMT.core import DataFrame


def read_data(filename, key=None, **kwargs):
    """General data reading routine.

    Used to read in data of an arbitrary file format. Correct reading routine is selected based on file extension.

    Parameters
    ----------
    filename : str
        Name of the file to be read including path

    Returns
    -------
    df       :  DMT.Dataframe()
        Dataframe that contains the cleaned up data contained in file filename
    """
    if not isinstance(filename, Path):
        filename = Path(filename)

    extension = filename.suffix.lower()

    if extension == ".mdm":
        df = read_mdm(str(filename))
    elif extension == ".elpa":
        df = read_elpa(filename)
    elif extension == ".crv":
        df = read_elpa(filename, header=3)
    elif extension == ".csv":
        df = read_csv(filename, **kwargs)
    elif extension == ".p":
        with filename.open("rb") as pickle_file:
            df = pickle.load(pickle_file)
    elif extension == ".h5":
        if key is None:
            key = "/df"
        df = read_hdf(filename, key)
    elif extension == ".feather":
        df = read_feather(filename)
    else:
        raise IOError(
            "Error: DMT can not open file "
            + filename
            + " since extension "
            + extension
            + " cannot be read."
        )

    # correct voltage names and make sure all potentials are available, account for imaginary numbers, e.g. columns R:S(1,1) and I:S(1,1) are stored as cmplx128 S(1,1)
    return df


def save_hdf(df, save_dir, filename):
    """Save the dataframe df into save_dir as filename.h5 .

    Parameters
    ----------
    df       :  pd.Dataframe()
        Pandas dataframe that shall be saved.
    save_dir :  str or os.Pathlike
        Directory where the file shall be saved
    filename : str
        Name of the file

    Returns
    -------
    boolean
        if saving was successfull returns True, else False
    """
    if not isinstance(save_dir, Path):
        save_dir = Path(save_dir)

    save_dir.mkdir(parents=True, exist_ok=True)

    return df.to_hdf(str(save_dir / (filename + ".h5")), key="df", mode="w")


def read_hdf(filename, key):
    """Read .h5 file and return the stored pandas dataframe from it.

    Parameters
    ----------
    filename      : str or os.Pathlike
        filename of .mdm file including path

    Returns
    -------
    dataframe  :  :class:`DMT.core.Dataframe`
        DMT DataFrame that stores the mdm data.
    """
    return pd.read_hdf(str(filename), key=key)


def read_mdm_blocks(block, n_row, n_col, n_iccap_vars, iccap_vars):
    data_raw = np.zeros([n_row, n_col + n_iccap_vars], dtype=np.float64)
    # read in raw data
    for i in range(n_row):
        data_raw[i, :n_col] = block[n_col * i : n_col * (i + 1)]
        data_raw[i, n_col : n_col + n_iccap_vars] = iccap_vars

    return data_raw


def read_mdm(filename):
    """Read .mdm file and generate a DMT dataframe from it.

    Parameters
    ----------
    filename : str or os.Pathlike
        filename of .mdm file including path.

    Returns
    -------
    :class:`DMT.core.Dataframe`
        DMT dataframe representing the mdm data.
    """
    block_data = []
    index_data = 0
    iccap_Vars = {}

    # open mdm for reading
    with open(filename, "r") as mdmfile:
        mdmtext = mdmfile.read()

    block_data = []
    for data in re.findall(r"^BEGIN_DB[\r\n]+([\s\S]*?)END_DB", mdmtext, re.MULTILINE):
        block_lines = data.splitlines()
        iccap_vars = {}
        for i, line in enumerate(block_lines):
            line = line.split()
            if "ICCAP_VAR" in line or "USER_VAR" in line:
                try:
                    iccap_vars[line[1]] = float(line[2])
                except ValueError:
                    raise ValueError(
                        "When trying to convert iccap_vars, ValueError. Could not convert to float."
                    )
                # pass
            else:
                if i == 0:
                    index_data = 0  # special case with no ICCAP_VARS
                else:
                    index_data = i + 1
                break
        # grab the cols
        cols = block_lines[index_data].replace("#", "").split()
        index_data += 1

        # get the raw data
        n_row = len(block_lines[index_data:])
        n_col = len(cols)
        n_iccap_vars = len(iccap_vars)

        block_str = " ".join(block_lines[index_data:])
        block_str = block_str.split()
        block = np.zeros(len(block_str))
        for i in range(len(block_str)):
            block[i] = float(block_str[i])

        data_raw = read_mdm_blocks(
            block,
            n_row,
            n_col,
            n_iccap_vars,
            np.array([value for key, value in iccap_vars.items()]),
        )
        # read in raw data
        # data_raw      =  np.zeros([n_row, n_col + n_iccap_vars ])
        # for i in range(n_row):
        #     data_raw[i,       :n_col             ]  =  block[ n_col*i : n_col*(i+1) ]
        #     data_raw[i,  n_col:n_col+n_iccap_vars]  =  [value  for key,value in iccap_vars.items()]

        cols.extend([key for key, value in iccap_vars.items()])
        block_data.append(data_raw)

    if block_data:
        all_data = np.concatenate(block_data)
        df = DataFrame(all_data, columns=cols)
        return df
    else:
        raise IOError("Failed to read from " + str(filename))


def read_elpa(filename, header=2):
    """Read .elpa file and generate a DMT dataframe from it.

    Parameters
    ----------
    filename : str
        filename of .elpa file including path.

    header   : integer
        Row number where actual data starts. Default=2 as is common for CEDIC in-house tools.

    Returns
    -------
    df       : :class:`DMT.core.Dataframe`
        DMT dataframe representing the elpa data.
    """
    # open file
    with open(filename, "r") as my_file:
        list_lines = my_file.readlines()

    # get column names
    list_lines = [line.strip() for line in list_lines]
    split_header = list_lines[1].split()
    columns = split_header[1:]  # names of columns

    try:  # catch device header splitting
        if int(split_header[0]) != len(columns):
            split_header2 = list_lines[2].split()
            try:
                float(split_header2[0])
                raise IOError(
                    "DMT->read_elpa: the number of columns and the length of the column names do not match. I tried to get more column names from the second line, but this failed."
                )
            except ValueError:
                columns = columns + split_header2
                header = 3 if header == 2 else header  # only tested for DEVICE
    except:
        pass

    # put all numeric values in large array and fill row by row taking n_data chunks
    # this omits the issue with line breaks produced e.g. by DEVICE
    list_lines = " ".join(list_lines[header:])
    list_lines = list_lines.split()
    list_lines = np.array([float(i) for i in list_lines])
    n_col = len(columns)
    n_row = len(list_lines) / len(columns)

    dummy = 1
    # check if n_row is an integer
    if n_row != round(n_row):
        warnings.warn(
            "DMT -> Data_reader: Encountered a weird number of rows in {:s} (n_row = {:f}).".format(
                filename, n_row
            )
        )

    n_row = int(n_row)  # but force it anyways!

    # fill the data into the 2-dimensional array data_raw
    data_raw = np.empty([n_row, n_col])
    for i in range(n_row):
        data_raw[i, :] = list_lines[n_col * i : n_col * (i + 1)]

    # initalize pd.Dataframe() and return it
    return DataFrame(data_raw, columns=columns)


def read_csv(filename, **kwargs):
    """Read .csv file and generate a DMT dataframe from it.

    Parameters
    ----------
    filename : str or os.Pathlike
        filename of .csv file including path.
    **kwargs
        kwargs are passed to pandas.read_csv

    Returns
    -------
    df       : :class:`DMT.core.Dataframe`
        DMT dataframe representing the csv data.
    """
    # ok, can just use standard pandas routine here. nice.
    df = pd.read_csv(str(filename), **kwargs)
    df.__class__ = DataFrame  # cast to DMT.DataFrame

    # work around for complex data in one column in the csv:
    # https://stackoverflow.com/a/18919965

    for col in df.columns:  # type: ignore
        if df[col].dtype == object:  # type: ignore
            try:
                df[col] = df[col].apply(lambda x: np.complex(x))  # type: ignore
            except AttributeError:
                # is there i instead of j ?
                df[col] = df[col].str.replace("i", "j").apply(lambda x: np.complex(x))  # type: ignore

    return df


def read_feather(filename, **kwargs):
    """Read .feather file and generate a DMT dataframe from it.

    Parameters
    ----------
    filename : str or os.Pathlike
        filename of .csv file including path.
    **kwargs
        kwargs are passed to pandas.read_csv

    Returns
    -------
    df       : :class:`DMT.core.Dataframe`
        DMT dataframe representing the csv data.
    """
    # ok, can just use standard pandas routine here. nice.
    df = pd.read_feather(str(filename), **kwargs)
    df.__class__ = DataFrame  # cast to DMT.DataFrame
    return df


def read_DEVICE_bin(filename):
    """Reads DEVICE binaries. Here the internal spacial data is saved. Returns a Dataframe.

    Parameters
    ----------
    filename : str
        Path to the file to read

    Returns
    -------
    df       : :class:`DMT.core.Dataframe`
        DMT dataframe representing the data.
    """

    def convert_to_int(ba_in):
        r"""Convert a byte array into an integer.

        Takes each byte from the input byte array and multiplies it with its "position"
        Example: ba_in = b'\xb4\t\x00\x00' => 00 00 09 B4 in hex
        First loop: char c = B4, should be multiplied with 1, (16*16)^0 = 1
        Second loop: char c = 09, should be multiplied with 256, (16*16)^1 = 256
        ...

        Parameters
        ----------
        ba_in : Byte array
            Read byte array from a binary file

        Returns
        -------
        int
            Calculated integer
        """
        return int(sum([c * pow(16 * 16, i_c) for i_c, c in enumerate(ba_in)]))

    # prepare struct unpack object
    obj_struct_f = struct.Struct("f")

    # first prepare dataframe
    with open(filename, mode="rb") as my_file:
        # length of one record
        ba_len_records = my_file.read(4)  # read an integer ( 4 Bytes)
        nr_len_records = convert_to_int(ba_len_records)  # [lrec]

        # read header
        # header is 76 Bytes long ( According to readdir.m )
        ba_header = my_file.read(76)
        _str_header = ba_header.decode()[3:75] + ba_header.decode()[0:2]

        # nr of axes
        # 1 for 1D, 2 for 2D, 3 for 3D
        nr_axes = convert_to_int(my_file.read(4))
        len_axes = [convert_to_int(my_file.read(4)) for i_axes in range(0, nr_axes)]
        len_axes_bytes = 4 + len(len_axes) * 4  # 4 bytes each, +8 from _nsum and _nend

        # next 2 are copied, I don't know their use
        _nsum = convert_to_int(my_file.read(4))
        _nend = convert_to_int(my_file.read(4))
        len_axes_bytes = len_axes_bytes + 8  # +8 from _nsum and _nend

        # read coordinates
        # construct indices: tuple of the coordinate permutations
        x = [obj_struct_f.unpack(my_file.read(4))[0] for i_x in range(0, len_axes[0])]
        name_columns = ["x"]
        len_axes_bytes = len_axes_bytes + len(x) * 4  # 4 bytes each, +8 from _nsum and _nend

        # y and z are only present in object if given in file
        if nr_axes >= 2:
            y = [obj_struct_f.unpack(my_file.read(4))[0] for i_y in range(0, len_axes[1])]
            name_columns.append("y")
            len_axes_bytes = len_axes_bytes + len(y) * 4  # 4 bytes each

            if nr_axes == 3:
                z = [obj_struct_f.unpack(my_file.read(4))[0] for i_z in range(0, len_axes[2])]
                name_columns.append("z")
                len_axes_bytes = len_axes_bytes + len(z) * 4  # 4 bytes each
                df_indices = []
                for z_a in z:
                    for y_a in y:
                        for x_a in x:
                            df_indices.append((x_a, y_a, z_a))
            else:
                df_indices = []
                for y_a in y:
                    for x_a in x:
                        df_indices.append((x_a, y_a))
        else:
            df_indices = x

    df = DataFrame(df_indices, columns=name_columns)

    # Read all records = saved variables over axis
    i_record = 0
    with open(filename, mode="rb") as my_file:
        while True:
            # jump to current record
            nr_jump = i_record * nr_len_records + 4
            my_file.seek(nr_jump, 0)  # 1 so relative to current position

            # read column name
            # header is 76 Bytes long ( According to readdir.m )
            ba_header = my_file.read(76)
            if len(ba_header) < 76:
                # here reading went wrong -> end of file!
                break
            col_name = ba_header.decode()[0:3].strip()

            # jump axis read
            my_file.seek(len_axes_bytes, 1)

            # read data in DEVICE sequence (for x,y,z index)
            read_vals = []
            if nr_axes == 1:
                # list of values
                for _i_x in range(0, len_axes[0]):
                    read_vals.append(obj_struct_f.unpack(my_file.read(4))[0])

            elif nr_axes == 2:
                # list readValues[i_y,i_x]
                for i_y in range(0, len_axes[1]):
                    # read_vals.append([])
                    for _i_x in range(0, len_axes[0]):
                        # read_vals[i_y].append(
                        read_vals.append(obj_struct_f.unpack(my_file.read(4))[0])

                # flatten the list of lists:
                # read_vals = list(chain(*read_vals))
            elif nr_axes == 3:
                # list readValues[i_z,i_y,i_x]
                for i_z in range(0, len_axes[2]):
                    read_vals.append([])
                    for i_y in range(0, len_axes[1]):
                        read_vals[i_z].append([])
                        for _i_x in range(0, len_axes[0]):
                            read_vals[i_z][i_y].append(obj_struct_f.unpack(my_file.read(4))[0])

                raise NotImplementedError(
                    "Needs testing how to flatten the list of lists of lists!"
                )

            df[col_name] = read_vals
            i_record = i_record + 1

    # check if pseudo 2D simulation
    if "y" in df.columns:
        if all((df["y"] == 0.0) | (df["y"] == df["y"].to_numpy()[-1])):
            # all 'y' are either 0.0 or 1.0 -> 1D simulation: cut off the df
            df = df[df["y"] == 0.0]
            del df["y"]

    return df


def read_ADS_bin(filename):
    """Reads a ADS raw data file into a DMT dataframe.

    Parameters
    ----------
    filename : str
        Path to the file to read

    Returns
    -------
    df       : :class:`DMT.core.Dataframe`
        DMT dataframe representing the data.
    """
    # read file
    with open(filename, "rb") as my_file:
        bin_content = my_file.readlines()

    # reverse order
    bin_content = bin_content[::-1]

    # binary 2 number
    # copy of sim_fe.read_raw_bin

    # content of file: simulations.name .type .var .data
    # 1st line: name
    _name = bin_content.pop()
    # 2nd line: date
    _date = bin_content.pop()

    # create output lists for both simulation types
    simulations = []
    simulations_dcop = []

    # prepare struct.Struct object for speed
    obj_struct_d = struct.Struct("d")

    # already get 1st line, for all others: rest after binary is this!
    line = bin_content.pop()
    while bin_content:
        sim = {}
        # analyze line

        # bugfix for letting line begin with Plotname
        # Use re, because it works in bytearrays
        pos = re.search(b"Plotname", line)
        while pos is None:
            # try to jump over this bias point :/
            # error prone because many things relay on same ordering as some reference...
            line = line + bin_content.pop()
            pos = re.search(b"Plotname", line)

        pos = pos.regs[0][0]
        line = line[pos:].decode()
        # type of simulation?
        if line[10:13] == "DC_":
            sim["name"] = line[18:]
            sim["type"] = line[10:17]
        else:
            sim["name"] = line[13:]
            sim["type"] = line[10:12]

        # next line is the format
        str_format = bin_content.pop().decode().strip()

        # number of variables and number of simulation points
        line = bin_content.pop()
        num_var = int(line[15:].strip())
        line = bin_content.pop()
        num_points = int(line[12:].strip())

        # Names of the columns
        sim["vars"] = []
        # in 1st column, there is "Variables:"
        for _i_var in range(num_var):
            line = bin_content.pop().strip().decode()
            str_splitted = line.split("\t")
            if str_splitted[0] == "Variables:":
                sim["vars"].append(str_splitted[2])
            else:
                sim["vars"].append(str_splitted[1])

        # remove "Binary:"
        line = bin_content.pop()[7:]

        # read the numbers according to the format
        if str_format == "Flags: real":
            if sim["type"] == "DC_DCOP":
                sim["data"] = []
                for _i_pt in range(num_points):
                    device_id = obj_struct_d.unpack_from(line)[0]
                    # check which type of simulation element is read
                    # at the moment _convert_CircuitElement_netlist in dut_ads sets the name to 3 characters!
                    str_splitted = sim["name"].split(".")
                    i_devicetype = len(str_splitted) - 1
                    if str_splitted[i_devicetype][0] == "R":
                        read_in = 3
                    elif str_splitted[i_devicetype][0] == "C":
                        read_in = 3
                    elif str_splitted[i_devicetype][0] == "S":
                        read_in = 3
                    elif str_splitted[i_devicetype][0] == "V":
                        read_in = 3
                    elif str_splitted[i_devicetype][0] == "I":
                        read_in = 3
                    else:
                        # VA-module
                        read_in = 3

                    line = bin_content.pop()
                    val = struct.unpack_from("{0}c".format(read_in), line)
                    val_decode = ""
                    for a in val:
                        try:
                            val_decode = val_decode + a.decode()
                        except UnicodeDecodeError:
                            break
                    sim["data_str"] = val_decode

                    data = np.zeros(num_var - 1)
                    data[0] = device_id
                    offset = read_in
                    for ndi_data in np.nditer(data[1:], op_flags=["writeonly"]):
                        while len(line) < offset + 8:
                            line = line + bin_content.pop()

                        ndi_data[...] = obj_struct_d.unpack_from(line, offset=offset)[0]
                        offset = offset + 8

                    sim["data"].append(data)
                    # old_line = line
                    # print('a: ' + str(line))
                    line = line[offset:]
                    # print('a: ' + str(line))

            else:
                data = np.zeros((num_var, num_points))
                offset = 0
                for ndi_data in np.nditer(data, order="F", op_flags=["writeonly"]):
                    while len(line) < offset + 8:
                        line = line + bin_content.pop()

                    ndi_data[...] = obj_struct_d.unpack_from(line, offset=offset)[0]
                    offset = offset + 8

                sim["data"] = data.transpose()
                # prepare for next iteration
                # old_line = line
                # print('b: ' + str(line))
                line = line[offset:]

        elif str_format == "Flags: complex":

            data = np.zeros((num_var, num_points), dtype=np.complex128)
            offset = 0

            # fortran order, 1h of bugfix Oo
            for ndi_data in np.nditer(data, order="F", op_flags=["writeonly"]):
                while len(line) < offset + 16:
                    line = line + bin_content.pop()

                a = obj_struct_d.unpack_from(line, offset=offset)[0]
                offset = offset + 8
                b = obj_struct_d.unpack_from(line, offset=offset)[0]
                offset = offset + 8

                ndi_data[...] = np.complex(a, b)

            sim["data"] = data.transpose()
            # old_line = line
            # print('c: ' + str(line))
            line = line[offset:]

        # save into output
        if sim["type"] == "DC_DCOP":
            simulations_dcop.append(sim)
        else:
            simulations.append(sim)

    return simulations, simulations_dcop


def save_elpa(fname, ELPA, cols, firstline):
    """Save data as a elpa file

    Parameters
    ----------
    fname : str or os.Pathlike
        Path to the file to save
    ELPA : object
        This parameter is passed to numpy.array as first parameter. Check this function for the supported types.
    cols : list[str]
        List of column names
    firstline : str
        First line (comment) for the file to create
    """
    # save xy data as elpa file
    # todo:does not work with pandas dataframes yet
    if not isinstance(fname, Path):
        fname = Path(fname)

    fname.parent.mkdir(parents=True, exist_ok=True)

    # make sure we have strings
    for i, col in enumerate(cols):
        cols[i] = str(col)

    with fname.open("w") as myfile:
        # --> write first line
        firstline = firstline.replace(chr(13), "")
        firstline = firstline.replace(chr(10), "")
        firstline = firstline + chr(13) + chr(10)
        myfile.write("{0:s}".format(firstline))

        # --> column line
        lc = len(cols)
        t = str(lc) + " "
        lc = len(t)
        if not cols == []:
            len_ = len(cols[0])
            t = t + chr(32) * (17 - len_ - lc) + cols[0] + " "
            for i in range(1, len(cols)):
                col = cols[i]
                len_ = len(col)
                if len_ > 17:
                    col = col[:17]
                    len_ = len(col)
                t = t + chr(32) * (17 - len_) + col + " "
            myfile.write("{0:s}\r\n".format(t[:-1]))

        # --> matrix
        ELPA = np.array(ELPA)

        if ELPA.shape[1] == len(cols):
            ELPA = np.transpose(ELPA)

        for ii in range(ELPA.shape[1]):
            for i in range(ELPA.shape[0]):
                if np.isnan(ELPA[i][ii]):
                    myfile.write("{0:17.9e} ".format(0))
                else:
                    myfile.write("{0:17.9e} ".format(ELPA[i][ii]))

            myfile.write("\r\n")


def read_tikz_file(filename, col_x=None, col_y=None):
    """Reads a tikz plot file into a DMT dataframe.

    Parameters
    ----------
    filename : str
        Path to the file to read
    col_x : str, optional
        Base name of the x columns, if not given, it is ugly extracted from the file
    col_y : str, optional
        Base name of the y columns, if not given, it is ugly extracted from the file

    Returns
    -------
    :class:`DMT.core.Dataframe`
        DMT dataframe representing the data.
    """
    # read file
    with open(filename, "r") as my_file:
        file_lines = my_file.readlines()

    # column base names
    for line in file_lines:
        if col_x is not None and col_y is not None:
            break

        if "xlabel" in line and col_x is None:
            col_x = re.search(r"\$(.+?)\$", line).group(1)
            if "/" in col_x:
                col_x = col_x.split("/")[0]
        elif "ylabel" in line and col_y is None:
            col_y = re.search(r"\$(.+?)\$", line).group(1)
            if "/" in col_y:
                col_y = col_y.split("/")[0]

    # column content
    data = {}
    for index, plotline in enumerate(
        re.finditer(
            r"\\addplot(.+?)};.+?(\\addlegendentry{(.+?)}\n|)", "".join(file_lines), flags=re.DOTALL
        )
    ):
        data_x = []
        data_y = []

        plt_lines = plotline.group(1).split("\n")

        for plt_line in plt_lines:
            match = re.match(r"([\d\.e-]+) ([\d\.e-]+)\\\\", plt_line)
            if match:
                data_x.append(float(match.group(1)))
                data_y.append(float(match.group(2)))

        sub_col = plotline.group(3)
        if sub_col is None:
            sub_col = str(index)

        data[col_x + sub_col] = np.array(data_x)
        data[col_y + sub_col] = np.array(data_y)

    return data



# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
name = "exceptions"


class Stopped(Exception):
    """Raised if the bounds are attempted to be changed such that self.low > self.up or such that the current value is not inside the new bounds."""


class Canceled(Exception):
    """Raised if the bounds are attempted to be changed such that self.low > self.up or such that the current value is not inside the new bounds."""


class ValueAtBoundsError(ValueError):
    """Raised when the input error is excluded."""


class SimulationUnsuccessful(Exception):
    """Raised if the output of a simulation is not valid."""


class SimulationFail(Exception):
    """Raised if the output of a simulation indicated a simulation failure."""


class ValueTooLargeError(ValueError):
    """Raised when the input value is too large."""


class ValueTooSmallError(ValueError):
    """Raised when the input value is too small."""


class ValueExcludedError(ValueError):
    """Raised when the input error is excluded."""


class ParaExistsError(LookupError):
    """Raised when a parameter is added to a composition in which a parameter with the same name already exists."""


class BoundsError(ValueError):
    """Raised if the bounds are attempted to be changed such that self.low > self.up or such that the current value is not inside the new bounds."""


class DataReferenceEmpty(Exception):
    """The xstep supplied init_data_reference_method did not generate reference data."""


class UnknownColumnError(Exception):
    """Raise if unknown columns in measurement data."""


class UnknownNodesError(Exception):
    """Raised if an known node is encountered during clean_names()."""


class NoOpenDeembeddingDut(Exception):
    """A suitable Open Deembedding Dut has not been found by DutLib.sort_duts."""


class NoShortDeembeddingDut(Exception):
    """A suitable Short Deembedding Dut has not been found by DutLib.sort_duts."""


class DeviceContactConfigError(Exception):
    """Raised by EvalTradica, if the given contact configuration (f.e. CBEBC) is unknown."""


class SpecifierNotKnown(IOError):
    """This specifier is unknown to DMT."""


class NanInfError(IOError):
    """Nan or Inf Error during calculation."""



# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import os
import re
import shutil
import _pickle as cpickle  # type: ignore

# import pickle as cpickle
import copy
import logging
from DMT.core import DatabaseManager, read_data, DataFrame, VAFile
from DMT.config import DATA_CONFIG
from pathlib import Path
import pandas as pd


class DutView(object):
    """DutView is the parent class of all DUTs in DMT.

    All simulation back-end classes must inherit from this class and overwrite the methods that raise NotImplemented errors.
    This ensures that all classes in DMT work independent of the concrete simulation back-end.

    Parameters
    ----------

    database_dir :  str or None
        The directory where all the databases of this project are saved. If None, DMT.config.DATA_CONFIG["directories"]["database"] is used.
    name       :  str
        The name of the DutView object, which is used as a prefix for saving the database and the pickled object.
    dut_type   : :class:`~DMT.core.dut_type.DutType`
        Type of the DutView object, represented by a DutType object (basically an Enumeration).
    nodes      : [string], optional
        List of names of the nodes of the DutView object. Only nodes that shall be present in the DutView's database are relevant.
        They are past in the given order to all df methods which need the small-signal parameters.
    copy_va_files : {False, True}, optional
    force      : {False, True}, optional
        If True, all data found in the duts database are deleted.
    separate_databases : {False, True}, optional
        If True, each simulation is saved in a separate database file and only the needed files are loaded.
        Usefull for large simulation files, like transient simulations. For small files, the lookup in the hard drive takes to long.
    list_copy : [], optional
        List of files or file contents to copy to the simulation directory as additional files.
    t_max : int, optional
        Maximum simulation duration for this dut. This value and the simulation controller value have to be exceeded.
    simulate_on_server : {False, True, None}, optional
        If not set, the config value is used.
    simulator_command : str
        Command to start the correct circuit simulator
    simulator_arguments : list[str]
        List of arguments for the simulator command, will be added one by one before the input file.
    technology     :  :class:`~DMT.core.technology.Technology`, optional
    width          :  float64, optional
        Width in m.
    length         :  float64, optional
        Length in m.
    nfinger        : int, optional
        Number of parallel emitter fingers
    contact_config : str, optional
        String describing the physical contact config for this device. The possible value depends on the technology, example: 'CBEBC'.
    inp_name   :  str
        The name of the input file to be generated.

    Attributes
    ----------
    database_dir :  str
        The directory where all the databases of this project are saved.
    name       :  str
        The name of the DutView object, which is used as a prefix for saving the database and the pickled object.
    manager    :  :class:`~DMT.core.database_manager.DatabaseManager`
        The DatabaseManager object of the DutView, used to interact with its database.
    sim_folder :  str
        Path to the simulation folder
    inp_name   :  str
        The name of the input file to be generated.
    save_dir   : string
        This is the directory were the DutView object will save its object and database.
    dut_dir    : string
        This is the directory were the DutView object will create its pickle file. This equals save_dir+'.h5'
    database_dir     : string
        This is the directory were the DutView object will create its database. This equals save_dir+'.h5'
    dut_type   : :class:`~DMT.core.dut_type.DutType`
        Type of the DutView object, represented by a DutType object (basically an Enumeration).
    nodes      : [string]
        List of names of the nodes of the DutView object. Only nodes that shall be present in the DutView's database are relevant.
        They are past in the given order to all df methods which need the small-signal parameters.
    ac_ports   : [string]
        List of ports for AC simulations. At each port mentioned here a AC source is applied. Defaults to the first 2 nodes.
    list_copy : [data]
        List of path or data to save to the simlation folder. If it is a path, the file is copied. If it is data, a file is created.

    Methods
    -------
    run_simulation(sweep)
        Start the simulation of this DutView object.
    prepare_simulation(sweep)
        Prepare a simulation folder and run make_input().
    make_input(sweep)
        Create the correct simulation input file from this Dut_view together with a Sweep object.
    get_hash()
        Return a unique hash for this Dut without considering the Sweep.
    save()
        Save this DutView as a pickled .p file to save_dir.
    __getstate__()
        This helper methods helps in the pickling process.
    __setstate__(state)
        This helper methods helps in the pickling process.
    load_dut(save_dir)
        Static class method. Loads a DutView object from a pickle file with full path save_dir.
    add_data(df, key, force=True)
        Add data into the DutView's database.
    get_data(key)
        Get the data stored in key from the DutView's database.
    check_existence_sweep(self, sweep)
        Returns True if the simulation corresponding to this DutView object in combination with Sweep object sweep has already been run.
    del_db()
        Delete the DutView's database.
    clean_all_data()
        Iterate through all keys in the DutView's database and try to clean the column names according to the internal DMT format.
    """

    def __init__(
        self,
        database_dir,
        name,
        dut_type,
        copy_va_files=True,
        force=False,
        separate_databases=False,
        list_copy=None,
        t_max=None,
        simulate_on_server=None,
        simulator_command="",
        simulator_arguments=None,
        technology=None,
        width=None,
        length=None,
        nfinger=None,
        contact_config=None,
        ac_ports=None,
        reference_node=None,
        nodes=None,
        inp_name=None,
        va_code_filter=None,
    ):
        self._copy_va_files = copy_va_files
        # files to be copied into simulation directory
        if list_copy is None:
            self.list_copy = []
        else:
            if not isinstance(list_copy, list):
                self.list_copy = [list_copy]
            else:
                self.list_copy = list_copy

        self._list_va_file_contents: list[VAFile] = []

        # attributes for data management
        self._separate_databases = separate_databases
        self._data = {}  # this is now hidden
        if database_dir is None:
            self._database_dir = DATA_CONFIG["directories"]["database"]
        else:
            if isinstance(database_dir, Path):
                self._database_dir = database_dir.expanduser().resolve()
            else:
                self._database_dir = Path(database_dir).expanduser().resolve()

        self.dut_type = dut_type

        if os.path.isabs(name):
            raise IOError("The DuT name must not be a valid absolute directory path!")
        self.name = name

        self.manager = DatabaseManager()
        if nodes is None:
            self.nodes = dut_type.get_nodes()
        else:
            if isinstance(nodes, str):
                self.nodes = nodes.split(",")
            else:
                self.nodes = nodes

        if ac_ports is None:
            self.ac_ports = self.nodes[0:2]  # ports for AC simulations
        else:
            self.ac_ports = ac_ports

        self.reference_node = reference_node
        if reference_node is None:
            raise IOError("Reference node not specified. DMT does not support this.")

        # simulation attributes
        self.inp_name = inp_name
        self.sim_command = simulator_command
        self.va_code_filter = va_code_filter

        if simulator_arguments is None:
            self.sim_args = []
        else:
            self.sim_args = simulator_arguments

        if force:
            # delete the saved database
            if self.save_dir is not None:
                self.del_db()
                self.del_dut()

        # try to reload this Dut if it already exists!
        if self.save_dir and not force:
            pickle = Path(self.dut_dir)
            if pickle.is_file():
                existing_dut = DutView.load_dut(self.dut_dir)
                self.__dict__.update(existing_dut.__dict__)

        self.sim_dir = DATA_CONFIG["directories"]["simulation"]

        self.technology = technology
        self.contact_config = contact_config
        self.width = width
        self.length = length
        self.nfinger = nfinger
        try:
            # if ( # was
            #     (width is not None)
            #     and (length is not None)
            #     and (not isinstance(length, tuple) and (not isinstance(width, tuple)))
            # ):
            self.perimeter = (self.width + self.length) * 2  # type: ignore
            self.area = self.width * self.length  # type: ignore
        except TypeError:
            # else:
            self.perimeter = None
            self.area = None

        # simulation attributes
        self.t_max = t_max
        if simulate_on_server is None:
            self.simulate_on_server = DATA_CONFIG["backend_remote"]
        else:
            self.simulate_on_server = simulate_on_server
        self.zip_result = True

        # in these attributes, store the name of another dut that has been used for deembedding...for documentation this is nice
        self.open_deembedded_with = "-"
        self.short_deembedded_with = "-"

    def prepare_simulation(self, sweep):
        """Creates a simulation folder, appends the sweep to the structure definition and creates a input file in the simulation folder.

        Parameters
        ----------
        sweep : Sweep
        """
        logging.info(
            "Preparing to simulate the Dut %s the sweep with the hash %s",
            self.get_hash(),
            sweep.get_hash(),
        )

        # set the simulation folder
        sim_folder = self.get_sim_folder(sweep)

        inp_content = self.make_input(sweep)

        try:
            sim_folder.mkdir(parents=True)
        except OSError:
            logging.debug("Folder already existed! Will be deleted")
            try:
                shutil.rmtree(str(sim_folder))
                sim_folder.mkdir(parents=True, exist_ok=True)
            except FileNotFoundError:
                raise OSError(
                    "Simulation folder exists but can not be deleted. Is the location mounted correctly? Hint: Shut down Windows correctly."
                )

        (sim_folder / self.inp_name).write_text(inp_content)
        if DATA_CONFIG["server"]["use_pbs"] and DATA_CONFIG["backend_remote"]:
            pbs_content = self.make_pbs(sweep)  # needs to be implemented by the DUT
            (sim_folder / "pbs_job").write_text(pbs_content)

        # copy va files?
        sim_folder = self.get_sim_folder(sweep)
        if self._copy_va_files:
            for vafile in self._list_va_file_contents:
                vafile.write_files(sim_folder, filter=self.va_code_filter)
        else:
            # write the files into a separate folder in simulations/va_files/HASH/
            va_files_dir = self.sim_dir / "VA_codes"
            for vafile in self._list_va_file_contents:
                dir_code = va_files_dir / vafile.get_tree_hash()
                if not dir_code.is_dir():
                    vafile.write_files(dir_code, filter=self.va_code_filter)

        for data_copy in self.list_copy:
            try:
                if os.path.isfile(data_copy):
                    filename = os.path.basename(data_copy)
                    shutil.copyfile(data_copy, os.path.join(sim_folder, filename))
                else:
                    # filename ??? Setting default file name... seems crazy here
                    (sim_folder / "datafile.tbl").write_text(data_copy)
            except TypeError:
                file_content = self._write_data_table(data_copy)
                file_name = sim_folder / "datafile.tbl"
                with file_name.open("a") as file_table:
                    file_table.write(file_content)

        str_log = " Prepared a simulation in the folder: \n{0:s}".format(str(sim_folder))
        print(str_log)
        logging.info(str_log)

    def get_start_sim_command(self):
        """Returns the command to start the simulation

        Returns
        -------
        str
        """
        return self.sim_command + " " + " ".join(self.sim_args) + " " + self.inp_name

    def _write_data_table(self, data):
        """Option to write a data table for a simulation. As the file type may depend on the simulator this is not implemented here!"""
        raise NotImplementedError("Has to be overwritten in the subclass.")

    def make_pbs(self, sweep):
        """Create a PBS script for the PBS job system. (See: https://albertsk.files.wordpress.com/2011/12/pbs.pdf)"""
        raise NotImplementedError("Not Implemented for this Dut Class!")

    def make_input(self, sweep):
        """Joins simulation header with a given Sweep object and returns it.

        Parameters
        ----------
        sweep : :class:`DMT.core.sweep.Sweep`
            Sweep specification according to the Sweep class.
        """
        raise NotImplementedError("DutView does not implement the concrete make_input method!")

    def get_hash(self):
        """Return a unique hash for this Dut without (!) considering the Sweep.

        Returns
        -------
        hash : float64
            Hash that corresponds to this DUT.
        """
        raise NotImplementedError("DutView does not implement the concrete get_hash method!")

    def import_output_data(self, sweep, delete_sim_results=False):
        """Read the output files that have been produced while simulating sweep and attach them to self.db.

        Parameters
        ----------
        sweep  :  :class:`DMT.core.sweep.Sweep`
            Sweep that has been simulated for the desired output files.
        delete_sim_results : {False, True}, optional
            If True, the simulation folder is deleted after reading.

        Raises
        ------
        NotImplementedError
            If the Dut is not a simulatable dut.
        IOError
            If the given sweep can not be read.
        """
        raise NotImplementedError(
            "DutView does not implement the concrete import_output_data method!"
        )

    def validate_simulation_successful(self, sweep):
        """Checks if the simulation of the given sweep was successful.

        Parameters
        ----------
        sweep  :  :class:`DMT.core.sweep.Sweep`
            Sweep that has been simulated.

        Raises
        ------
        NotImplementedError
            If the Dut is not a simulatable dut.
        SimulationUnsuccessful
            If the simulation output is not valid.
        FileNotFoundError
            If the sim log file does not exist.
        """
        raise NotImplementedError(
            "DutView does not implement the concrete import_output_data method!"
        )

    @property
    def save_dir(self):
        if self.get_hash():
            return self.database_dir / (self.name + "_hash_" + str(self.get_hash()))
        else:
            return self.database_dir / self.name

    @property
    def database_dir(self):
        return self._database_dir

    @database_dir.setter
    def database_dir(self, new_dir):
        """Make sure data is loaded before changing save location of this DutView."""
        if not self._data:
            self.load_db()

        self._database_dir = Path(new_dir)

    def get_db_dir(self, name="db"):
        """Returns the name for a db, either use 'db' for regular behavior or use 'sweep.get_hash()' for a db per sweep."""
        return self.save_dir / (name + ".h5")

    @property
    def dut_dir(self):
        return self.save_dir / "dut.p"

    @property
    def data(self):
        """data is a property to ensure loading before usage.

        As _data is a dict, the getter is also called before setting. So the separate setter is not necessary.
        If someone tries to set dut.data an attribute error occurs, but setting dict entries is possible directly.

        """
        if not self._data:  # empty dict evaluates to false
            self.load_db()

        return self._data

    def get_sim_folder(self, sweep):
        """Returns the simulation folder of the given sweep

        Parameters
        ----------
        sweep  :  Sweep
            Sweep object that corresponds to the folder.

        Returns
        -------
        str
            Path to the simulationfolder
        """
        return (
            self.sim_dir
            / (self.name + str(self.get_hash()))
            / (sweep.name + "_" + sweep.get_hash())
        )

    def delete_sim_results(self, sweep, ignore_errors=False):
        """Deletes the simulation results of the given sweep.

        Parameters
        ----------
        sweep  :  Sweep
            Sweep object that corresponds to the folder.
        ignore_errors : {False, True}, optional
            Passed through to :meth:`shutil.rmtree`. If ignore_errors is true, errors resulting from failed removals will be ignored.
        """
        sim_folder = self.get_sim_folder(sweep)
        shutil.rmtree(sim_folder, ignore_errors=ignore_errors)

        logging.info(
            "The simulation data of the sweep %s with the hash %s was deleted",
            sweep.name,
            sweep.get_hash(),
        )

    def save(self):
        """Save this DutView as a pickled .p file and the also the data into the database on the hard drive."""
        self.save_dir.mkdir(parents=True, exist_ok=True)
        self._database_dir = self._database_dir.resolve()  # convert to absolute path

        self.save_db()
        with open(self.dut_dir, "wb") as handle:
            cpickle.dump(self, handle)
        # with warnings.catch_warnings():
        # warnings.simplefilter('ignore', tb.NaturalNameWarning)

    def __getstate__(self):
        """Return state values to be pickled. Implemented according `to <https://www.ibm.com/developerworks/library/l-pypers/index.html>`_ .
        Notes
        -----
        ..todo:
            iterate through all properties and throw away the HDFStore objects.
        """
        d = copy.copy(self.__dict__)

        if "_data" in d:
            del d["_data"]

        return d

    def __setstate__(self, state):
        """Return state values to be pickled. Implemented according `to <https://www.ibm.com/developerworks/library/l-pypers/index.html>`_ ."""
        self.__dict__ = state  # pylint: disable=attribute-defined-outside-init
        self.__dict__["_data"] = {}

    @staticmethod
    def load_dut(file_dut):
        """Static class method. Loads a DutView object from a pickle file with full path save_dir.

        Parameters
        ----------
        file_dut  :  str or os.Pathlike
            Path to the pickled DutView object that shall be loaded.

        Returns
        -------
        obj  :  DutView()
            Loaded object from the pickle file.
        """
        if not isinstance(file_dut, Path):
            file_dut = Path(file_dut)

        with file_dut.open(mode="rb") as handle:
            dut = cpickle.load(handle)

        # check the dirs:
        # obtain database_dir from file_dut
        database_dir = file_dut.parent.parent

        # check if equal, if not -> changed machine -> broken absolute path
        if database_dir != dut._database_dir:  # pylint: disable=protected-access
            dut._database_dir = database_dir  # pylint: disable=protected-access

        if not os.path.exists(dut.sim_dir):
            dut.sim_dir = DATA_CONFIG["directories"]["simulation"]

        return dut

    def add_data(self, data, key=None, force=True, **kwargs):
        """Add a measurement or simulation data to the DutView's data.

        Parameters
        ----------
        data   :   DataFrame, sweep or str
            If DataFrame: Directly added to the data using the given key.
            If str: Full path to the file that shall be added to the database
            if sweep: Simulated sweep to import.
        key    :   string, optional
            Key that shall be used in the database to save the data.
        force  :  bool, optional
            Default=True. If = True, the data is added even if it already exists.
        """
        if key is None:
            try:
                key = self.get_sweep_key(data)
            except AttributeError:
                pass

        if key in self.data.keys():
            if force:
                logging.info(
                    "DMT -> DutView -> add_data(): Removed a dataframe with key %s since it was already existent in dut.data.",
                    key,
                )
                self.remove_data(key)
            else:
                logging.info(
                    "DMT -> DutView -> add_data(): Skipped a dataframe with key %s since it was already existent in dut.data.",
                    key,
                )
                return

        try:
            self.data[key] = read_data(data, **kwargs)
        except (OSError, IOError, TypeError) as _error:
            # could not read file using the general read data!
            if isinstance(data, DataFrame) or isinstance(data, pd.DataFrame):
                # it is a regular dataframe
                # prevents pandas bug with non-unique columns:
                if not data.columns.is_unique:
                    data = data.loc[:, ~data.columns.duplicated()]
                self.data[key] = data
            else:
                # simulation valid?
                self.validate_simulation_successful(data)
                # try special import
                self.import_output_data(data)

        logging.info("DMT -> DutView -> add_data(): Added a dataframe with key %s to the dut.", key)

    def remove_data(self, key):
        """Remove a measurement or simulation dataframe from the DutView's data.

        Parameters
        ----------
        key    :   string
            Key that shall be removed from the database.
        """
        del self.data[key]

    def get_data(self, key="iv", sweep=None):
        """Return data stored in the DutView's data.

        | One needs to specify either:
        | - key      : The data stored under the path key in the dut's database is returned.
        | - sweep    : Get the data from this sweep. If none, key must be a valid key for the database.
        | - sweep+key: Return the data stored as self.get_sweep_key(sweep)+'/'+key from the dut's database.

        Parameters
        ----------
        key    :  str, optional, {'iv'}
            The key of the data in the DutView's database that shall be retrieved.
        sweep  :  :class:`DMT.core.sweep.Sweep`, optional
            The sweep whose data shall be looked for.
        """
        if self._separate_databases:
            if sweep is None:
                raise IOError("If you do not know the sweep, just use dut.data[key]..")

            key = self.join_key(self.get_sweep_key(sweep), key)
            if key not in self._data.keys():
                self.load_db(sweep)

        else:
            if sweep is not None:
                key = self.join_key(self.get_sweep_key(sweep), key)

            if not self._data:
                self.load_db()

        try:
            return self._data[key]
        except KeyError:
            ## missed the starting '/' ?
            return self._data[self.join_key("", key)]

    def del_db(self):
        """Delete the DutView's complete database."""
        # iterate over the folder in case of self.separate_databases
        for root, _dirs, files in os.walk(self.save_dir):
            for my_file in files:
                if my_file.endswith(".h5"):
                    self.manager.del_db(root + "/" + my_file)
        logging.info("DMT -> DutView -> del_data(): Deleted a complete database.")

    def del_dut(self):
        """Delete the DutView's pickled file."""
        if os.path.exists(self.dut_dir):
            os.remove(self.dut_dir)

        logging.info("DMT -> DutView -> del_dut(): Deleted a pickled DutView object.")

    def clean_data(self, fallback=None, **kwargs):
        """Clean the dataframe columns of the DataFrame objects in this DutMeas objects database.

        Parameters
        ----------
        fallback : dict, optional
            Is used to update self.nodes_fallback for this special key to clean.

        """
        if fallback is None:
            fallback = {}
        for key, df in self.data.items():
            try:
                df = df.clean_data(
                    self.nodes,
                    self.reference_node,
                    fallback=fallback,
                    ac_ports=self.ac_ports,
                    **kwargs,
                )
            except AttributeError:
                df.__class__ = DataFrame  # cast to DMT.DataFrame
                df = df.clean_data(
                    self.nodes,
                    self.reference_node,
                    fallback=fallback,
                    ac_ports=self.ac_ports,
                    **kwargs,
                )

            self.data[key] = df

    def save_db(self):
        """Write a database for this dut. If it already exists it is overwritten. Does NOT save all keys starting with '_'"""
        if not self._data:
            return  # nothing to do here

        if self._separate_databases:
            # find all sweeps in self.data
            sweep_keys = []
            for key in self._data.keys():
                ## key is equal except for the last part -> same sweep
                sweep_key = self.join_key(*self.split_key(key)[0:-1])
                if sweep_key not in sweep_keys:
                    sweep_keys.append(sweep_key)

            for sweep_key in sweep_keys:
                data_to_save = {}
                for key, data in self._data.items():
                    if (
                        not key.startswith("_")
                        and self.join_key(*self.split_key(key)[0:-1]) == sweep_key
                    ):
                        data_to_save[key] = data

                self.manager.save_db(self.get_db_dir(sweep_key), data_to_save)
        else:
            data_to_save = {}
            for key, data in self._data.items():
                if not key.startswith("_"):
                    data_to_save[key] = data

            self.manager.save_db(self.get_db_dir(), data_to_save)

    def load_db(self, sweep=None):
        """Load saved data from the database into the object.

        If the complete dictionary is reloaded, here it is set directly to self._data.
        if separate databases is activated it loads only the given sweep (and it has to be given then!)

        Parameters
        ----------
        sweep : Sweep or str, optional
            In case of separate databases, the sweep must be given either directly or as string name
        """
        if self._separate_databases:
            if sweep is None:
                return

            if isinstance(sweep, str):
                name = sweep
            else:
                name = self.get_sweep_key(sweep)

            try:
                self._data.update(self.manager.load_db(self.get_db_dir(name)))
            except FileNotFoundError:
                pass
        else:
            try:
                self._data = self.manager.load_db(self.get_db_dir())
            except FileNotFoundError:
                pass

    def check_existence_sweep(self, sweep):
        """Return true, if the combination dut+sweep has already been simulated.

        If self.data is None, the database is loaded first.

        Parameters
        ----------
        sweep  : :class:`DMT.core.sweep.Sweep`
            Sweep object that can return a hash.

        Returns
        -------
        existence  :  bool
            Is True if the combination dut+sweep has already been simulated.

        """
        key_sweep = self.get_sweep_key(sweep)
        if self._separate_databases:
            if not any(key_sweep in data_key for data_key in self.data.keys()):  # needs full check
                self.load_db(sweep)
        else:
            if not self._data:  # easy check
                self.load_db()

        return any(key_sweep in data_key for data_key in self.data.keys())

    def join_key(self, *parts_key):
        """Joins the parts of the key into one key for self.dict

        Parameters
        ----------
        parts_key : iterable

        Returns
        -------
        str
            parts_key[0]/parts_key[1]/.../parts_key[-1]
        """
        return "/".join(parts_key)

    def split_key(self, key):
        """Splits the key up in its parts.

        Parameters
        ----------
        key : str

        Returns
        -------
        list[str]
            key splitted at '/'

        """
        return key.split("/")

    def get_sweep_key(self, sweep):
        """Key for the dict in dut.data.

        Parameters
        ----------
        sweep : :class:`~DMT.core.sweep.Sweep` or str
            Either the key for the given sweep or if it is a string, directly the string

        Returns
        -------
        key : str
        """
        try:
            return self.join_key(sweep.get_temperature(), sweep.name + "_" + sweep.get_hash())
        except AttributeError:
            # given sweep parameter does not have a get_hash(), is it a string?!?
            return sweep

    def get_key_temperature(self, key):
        """Function that returns the temperature of a given data key. Overwrite this if the measurements differ from the default DMT naming.

        Default naming is:

        * Single temperature: "Txxx.xxK"
        * List of temperatures: "T(xxx.xx,yyy.yy,...)K"
        * Range of temperatures: "T[xxx.xx-sss.ss-yyy.yy]K", s is the step

        Parameters
        ----------
        key : str
            Key that shall be evaluated.

        Returns
        -------
        temp : float
            The temperature at which the measurement "key" has been conducted in Kelvin.
        """
        key_parts = self.split_key(key)

        # first check the best and most usefull way for a single temperature
        for key_part in key_parts:
            re_temp = re.search(r"T([0-9p\.]+)K", key_part)
            if re_temp:
                try:
                    # always replace "p" with ".", if it is already with ".", it doesnt matter
                    return round(float(re_temp.group(1).replace("p", ".")), 3)
                except ValueError:
                    # if a value error in the except clause happens, try the next key part.
                    pass

        # check for list
        for key_part in key_parts:
            re_temp = re.search(r"T[(](.+)[)]K", key_part)
            if re_temp:
                try:
                    str_temps = re_temp.group(1).split(",")
                    return [round(float(str_temp), 3) for str_temp in str_temps]
                except ValueError:
                    # if a value error in the except clause happens, try the next key part.
                    pass

        # check for range
        for key_part in key_parts:
            re_temp = re.search(r"T\[(.+)\)K", key_part)
            if re_temp:
                try:
                    str_temps = re_temp.group(1).split("-")
                    return (round(float(str_temp), 3) for str_temp in str_temps)
                except ValueError:
                    # if a value error in the except clause happens, try the next key part.
                    pass

        # alternative:
        for key_part in key_parts:
            if key_part.startswith("T"):
                try:
                    return round(float(key_part[1:].replace("p", ".")), 3)
                except ValueError:
                    # if a value error in the except clause happens, try the next key part.
                    pass

        # finally as a last escape: direct conversion :(
        for key_part in key_parts:
            try:
                return round(float(key_part.replace("p", ".")), 3)
            except ValueError:
                # if a value error in the except clause happens, try the next key part.
                pass

        raise NameError(
            "DMT -> DutMeas -> get_key_temperature: Was not able to extract the temperature from the key: "
            + key
            + "."
        )



""" DMT interaction with pypi
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import requests
import re
import argparse


def get_file_idx(release, pattern=".whl"):
    return next(file_a for file_a in release if re.search(pattern, file_a["filename"]))


def get_pypi_url(package="DMT-core", version=None, pattern=".whl"):
    package = requests.get(f"https://pypi.python.org/pypi/{package:s}/json").json()
    if version is None:
        version = max(package["releases"].keys())
    # ... check compatibility
    file_to_get = next(
        file for file in package["releases"][version] if re.search(pattern, file["filename"])
    )
    return file_to_get["url"]


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Obtain the link to a file at pypi.org")
    parser.add_argument(
        "--package",
        type=str,
        help="Package to query (default: %(default)s).",
        default="DMT-core",
        required=False,
    )
    parser.add_argument(
        "--version",
        type=str,
        help="Version of the query. If None, the newest version is used (default: %(default)s).",
        default=None,
        required=False,
    )
    parser.add_argument(
        "--pattern",
        type=str,
        help="Pattern to scan the different files of the release (default: %(default)s).",
        default=".whl",
        required=False,
    )
    args = parser.parse_args()

    file_link = get_pypi_url(package=args.package, version=args.version, pattern=args.pattern)
    print(file_link)



""" Helpers to integrate VerilogAE into DMT
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import warnings
from inspect import signature
from collections import OrderedDict

# VAE models known to DMT
HICUM_L2 = "HICUM_L2"
SGP = "SGP"
HICUM_L0 = "HICUM_L0"


def get_param_list(meq_function, all_parameters=False, info=None):
    """Returns a list with the McParameter names for the given callable in correct order, assuming that only the top-most function is not from VerilogAE

    Parameters
    ----------
    meq_function : function
        Function of the model equation which shall be used.
    all_parameters : {False, True}, optional
        If True, the independent_vars are ignored and the full parameter list is returned.

    Returns
    -------
    params : list
        List of parameters for this function
    """
    if info is None:
        info = {}

    try:
        sig = signature(meq_function)
        func_params = list(sig.parameters)

        # only kwargs : https://stackoverflow.com/questions/196960/can-you-list-the-keyword-arguments-a-function-receives
        func_params = [p.name for p in sig.parameters.values() if (p.kind == p.KEYWORD_ONLY)]
    except TypeError:
        # if meq_function is directly a verilogae function
        func_params = meq_function.parameters

    for attr in ["depends", "depends_optional", "independent_vars"]:
        if not attr in info:
            info[attr] = tuple()

    if not isinstance(info["depends"], tuple):
        raise NotImplementedError("MUST be tuple! Error in " + meq_function.__name__ + "_info")
    if not isinstance(info["depends_optional"], tuple):
        raise NotImplementedError("MUST be tuple! Error in " + meq_function.__name__ + "_info")
    if not isinstance(info["independent_vars"], tuple):
        raise NotImplementedError("MUST be tuple! Error in " + meq_function.__name__ + "_info")

    for dependence in info["depends"]:
        if isinstance(dependence, str):
            func_params.append(dependence)
        else:
            try:
                func_params += dependence.parameters
            except AttributeError:
                func_params += get_param_list(dependence, all_parameters=all_parameters)

    # unique it!
    func_params = list(OrderedDict.fromkeys(func_params))

    if all_parameters:
        func_params = list(OrderedDict.fromkeys(func_params))
        return func_params

    # delete the parameters which are independent and without the opti_params
    params = []
    for param in func_params:
        if not param in info["independent_vars"]:
            params.append(param)

    return params


def get_dmt_model(vae_module, model_type, version):
    """Retrieving a DMT fitting model for XSteps from a VAE compiled VA-Code.

    This function also adds the needed attributes.

    Parameters
    ----------
    vae_module : module
        VAE compiled and installed VA-Code
    model_type : {'HICUM_L0', 'HICUM_L2'}
        Currently supporting only the two HICUM levels
    version : float
        Version of the compiled va-code

    Returns
    -------
    model : module
        The VAE module with added attributes

    """
    warnings.warn(
        "get_DMT_model is deprecated and will be removed in future major releases.\n"
        + "Directly pass the MCard and optionally the special VAE module to the XSteps!",
        category=DeprecationWarning,
    )
    model = vae_module

    if model_type not in [HICUM_L0, HICUM_L2, SGP]:
        raise IOError("DMT->VerilogAE: Currently only 'HICUM' is supported by VAE+DMT.")

    # temprary fix
    class VaeStub:
        pass

    model_patched = VaeStub()
    model_patched.functions = model.functions
    model_patched.modelcard = model.modelcard
    model = model_patched
    # temprary fix

    model.model_type = model_type
    model.version = version
    if "HICUM" in model_type:
        if "L0" in model_type:
            model.hicum_level = 0
        elif "L2" in model_type:
            model.hicum_level = 2
        else:
            raise IOError("DMT->VerilogAE: The HICUM level must be inside the model_type!")
    elif "SGP" in model_type:
        model.hicum_level = 0  # Markus: why do we need that
    else:
        raise IOError("DMT->VerilogAE: Currently only 'HICUM' is supported")

    return model



""" The interface to the Xyce circuit simulator from Sandia
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import copy
import logging
import re
import subprocess
import warnings
import numpy as np
import pandas as pd
import time
from pathlib import Path

from DMT.config import DATA_CONFIG
from DMT.core import (
    DutCircuit,
    DutType,
    McParameterCollection,
    MCard,
    Sweep,
    SweepDef,
    specifiers,
    sub_specifiers,
    SpecifierStr,
    DataFrame,
    create_md5_hash,
    constants,
    print_progress_bar,
)
from DMT.core.circuit import (
    Circuit,
    RESISTANCE,
    INDUCTANCE,
    CAPACITANCE,
    VOLTAGE,
    CURRENT,
    SHORT,
    HICUML2_HBT,
    SGP_BJT,
)
from DMT.exceptions import SimulationUnsuccessful, SimulationFail


class DutXyce(DutCircuit):
    """Class to interface the Xyce circuit simulator"""

    def __init__(
        self,
        database_dir,
        dut_type,
        input_circuit,
        name="xyce_",
        simulator_command=None,
        simulator_arguments=None,
        simulator_options=None,
        build_xyce_plugin_command="buildxyceplugin",
        **kwargs,
    ):
        if simulator_command is None:
            simulator_command = DATA_CONFIG["commands"]["XYCE"]

        default_options = {}
        if simulator_options is not None:
            default_options.update(simulator_options)
        simulator_options = default_options

        if simulator_arguments is None:
            simulator_arguments = []

        self.build_xyce_plugin_command = build_xyce_plugin_command
        self._va_plugins_to_compile = []

        super().__init__(
            database_dir,
            name,
            dut_type,
            input_circuit,
            simulator_command=simulator_command,
            simulator_options=simulator_options,
            simulator_arguments=simulator_arguments,
            inp_name="xyce_circuit.cir",
            **kwargs,
        )

        if self.simulate_on_server:
            raise NotImplementedError(
                "Running Xyce on a Server is not implemented at the moment.",
                "It could work for non-VA Duts, but needs testing.",
                "Verilog Xyce does not work for sure.",
            )

    def create_inp_header(self, inp_circuit):
        """Creates the input header of the given circuit description and returns it.

        Parameters
        ----------
        input : DMT.core.Circuit or DMT.core.MCard
            If a MCard is given, the circuit is created from it.

        Returns
        -------
        netlist : str
        """
        va_from_model_build_in = False
        if isinstance(inp_circuit, MCard) or isinstance(inp_circuit, McParameterCollection):
            # save the modelcard, in case it was set inderectly via the input header!
            self._modelcard = copy.deepcopy(inp_circuit)
            # get the circuit for netlist generation

            ## DIRTY MARKUS FIX NETLIST VARIABLES ##############
            ## now fixed directly in the default circuit
            ####################################################

            # in case a standard circuit is used, this is the real input circuit
            self._inp_circuit = inp_circuit.get_circuit(**self.get_circuit_arguments)  # type: ignore
            if (
                "use_build_in" in self.get_circuit_arguments
                and self.get_circuit_arguments["use_build_in"]
            ):
                va_from_model_build_in = True
        elif isinstance(inp_circuit, Circuit):
            self._modelcard = None
            self._inp_circuit = copy.deepcopy(inp_circuit)
        else:
            raise OSError(
                "For Xyce circuits netlist generation is only possible from object of class DMT.classes.Circuit. Passed "
                + str(inp_circuit)
                + "."
            )

        str_netlist = (
            "DMT Xyce simulation\n"
            # "simulator1Options options "
            # + self._convert_dict_to_inp_line(self.simulator_options)
            # + "\n\n"
        )

        # is a modelcard inside the netlist?
        list_va_files = list()
        for element in self._inp_circuit.netlist:
            try:
                list_va_files.append(element.parameters.va_codes)
            except AttributeError:
                # element does not have a va_file.
                pass

        if not va_from_model_build_in:
            # load va file plugins and mark missing ones -> they are compiled later:
            va_plugins = []
            self._va_plugins_to_compile = []  # reset the plugins
            for vafile in list_va_files:
                if vafile is None:
                    continue

                if self._copy_va_files:
                    self._list_va_file_contents.append(vafile)

                # hash file content
                va_hash = vafile.get_tree_hash()
                # check if plugin is already compiled and in the folder
                path_va_hash = self.sim_dir / "xyce_plugins" / (va_hash + ".so")

                if not path_va_hash.is_file():
                    # if not: add to "to_compile"
                    self._va_plugins_to_compile.append((path_va_hash, vafile))

                va_plugins.append(path_va_hash)

            # add to simulator arguments
            if va_plugins:
                try:
                    i_plugin = self.sim_args.index("-plugin")
                    self.sim_args[i_plugin + 1] = ",".join(str(va_path) for va_path in va_plugins)
                except ValueError:
                    self.sim_args.append("-plugin")
                    self.sim_args.append(",".join(str(va_path) for va_path in va_plugins))

        str_netlist += "\n* Netlist\n"

        # add elements:
        for index, element in enumerate(self._inp_circuit.netlist):
            if isinstance(element, str):
                if not "include" in element:
                    str_netlist += ".PARAM " + element + "\n"
                else:
                    raise NotImplementedError("Need a test case for this ?!?")
                    str_netlist += element + "\n"

            else:
                str_netlist = str_netlist + self._convert_CircuitElement_netlist(element, index)

        logging.info("Successfully created input header of dut %s!", self.name)
        logging.debug("Content:\n%s", str_netlist)

        return str_netlist

    def _convert_CircuitElement_netlist(self, circuit_element, index):
        """Transforms a :class:`~DMT.classes.circuit.CircuitElement` into a string fitting for ADS.

        Parameters
        ----------
        circuit_element : :class:`DMT.core.CircuitElement`
            CircuitElement to transform
        index : int
            Counting index of the element -> to count the different models and stuff and make sure they are different

        Returns
        -------
        str
            Netlist line
        """
        converter = {
            RESISTANCE: "R",
            CAPACITANCE: "C",
            INDUCTANCE: "I",
        }

        if circuit_element.element_type == SHORT:
            return (
                "R"
                + circuit_element.name
                + " "
                + " ".join(circuit_element.contact_nodes)
                + " R=0 \n"
            )
        elif circuit_element.element_type == VOLTAGE:
            # CircuitElement( VOLTAGE, 'V_B', ['n_BX', '0'], parameters=[('Vdc','V_B'), ('Vac','1')])
            dc = next(para for para in circuit_element.parameters if para[0] == "Vdc")[1]
            try:
                float_dc = float(dc)
                parameters = "DC " + dc + " V"
            except ValueError:
                # if dc contaíns a string -> it is a parameter and hence no unit needed but braces...
                parameters = "DC {" + dc + "}"
            try:
                ac = next(para for para in circuit_element.parameters if para[0] == "Vac")[1]
                try:
                    float_ac = float(ac)
                    ac = ac + "V"
                except ValueError:
                    # if dc contaíns a string -> it is a parameter and hence no unit needed but braces...
                    ac = "{" + ac + "}"
                phase = str(np.rad2deg(0))  # just to remember this
                parameters += " AC " + ac + " " + phase
            except StopIteration:
                pass
            return (
                "V"
                + circuit_element.name
                + " "
                + " ".join(circuit_element.contact_nodes)
                + " "
                + parameters
                + " \n"
            )
        elif circuit_element.element_type == CURRENT:
            # CircuitElement( CURRENT, 'ISB', ['0', 'n_BX'], parameters=[('Idc','IB'), ('Iac','1')])

            dc = next(para for para in circuit_element.parameters if para[0] == "Idc")[1]
            parameters = "DC " + dc + "V"
            try:
                ac = next(para for para in circuit_element.parameters if para[0] == "Iac")[1]
                phase = str(np.rad2deg(0))  # just to remember this
                parameters += " AC " + ac + "V " + phase
            except StopIteration:
                pass

            return (
                "I"
                + circuit_element.name
                + " "
                + " ".join(circuit_element.contact_nodes)
                + " "
                + parameters
                + " \n"
            )
        elif circuit_element.element_type == HICUML2_HBT:
            # a model with the parameters
            # the model name
            model_name = circuit_element.parameters.default_subckt_name + "_{:d}".format(index)

            # the instance
            str_netlist = (
                "Q"
                + circuit_element.name
                + " "
                + " ".join(circuit_element.contact_nodes)
                + " "
                + model_name
                + "\n"
            )

            # and now the model with the parameters
            str_netlist += ".MODEL " + model_name + " NPN ( LEVEL=234"
            for para in circuit_element.parameters.iter_alphabetical():
                if not para.name.startswith("_"):  # do not use parameters that start with _
                    if para.val_type == int:
                        str_netlist += " {0:s} = {0:d}".format(para)
                    else:
                        str_netlist += " {0:s} = {0:10.10e}".format(para)

            return str_netlist + " )\n"
        elif circuit_element.element_type == SGP_BJT:
            # a model with the parameters
            # the model name
            model_name = circuit_element.parameters.default_subckt_name + "_{:d}".format(index)

            # the instance
            str_netlist = (
                "Q"
                + circuit_element.name
                + " "
                + " ".join(circuit_element.contact_nodes)
                + " "
                + model_name
                + "\n"
            )

            # and now the model with the parameters
            str_netlist += ".MODEL " + model_name + " NPN ( LEVEL=1"
            for para in circuit_element.parameters.iter_alphabetical():
                if not para.name.startswith("_"):  # do not use parameters that start with _
                    if para.val_type == int:
                        str_netlist += " {0:s} = {0:d}".format(para)
                    else:
                        str_netlist += " {0:s} = {0:10.10e}".format(para)

            return str_netlist + " )\n"
        else:
            try:
                # resistors, capacitors and inductors
                str_netlist = (
                    converter[circuit_element.element_type]
                    + circuit_element.name
                    + " "
                    + " ".join(circuit_element.contact_nodes)
                    + " "
                )
                for para, value in circuit_element.parameters:
                    str_netlist += para.upper() + "=" + value + " "

                return str_netlist + " \n"

            except KeyError:
                # special devices like the transistor
                model_name = circuit_element.parameters.default_subckt_name + "_{:d}".format(index)
                str_netlist = (
                    "Y"
                    + circuit_element.element_type
                    + " "
                    + circuit_element.name
                    + " "
                    + " ".join(circuit_element.contact_nodes)
                    + " "
                    + model_name
                    + "\n"
                )
                # this try-except works only if the parameters are a DMT modelcard
                str_temp = ".MODEL " + model_name + " " + circuit_element.element_type
                try:
                    for para in circuit_element.parameters.iter_alphabetical():
                        if not para.name.startswith("_"):  # do not use parameters that start with _
                            if para.val_type == int:
                                str_temp += " {0:s} = {0:d}".format(para)
                            else:
                                str_temp += " {0:s} = {0:10.10e}".format(para)
                except AttributeError:
                    for para, value in circuit_element.parameters:
                        str_temp += " " + para + "=" + value

            return str_netlist + str_temp + "\n"

    def make_input(self, sweep):
        """Creates the sweep definition for the given sweep

        Parameters
        ----------
        sweep : Sweep

        Returns
        -------
        str
            header with added bias definitions
        """
        # first make copy of sweep, ensuring that the actual sweep object is not changed
        tmp_sweep = copy.deepcopy(sweep)
        # sweepdefs = copy.deepcopy(tmp_sweep.sweepdef)

        # compile plugins if needed and also add to arguments
        if self._va_plugins_to_compile:
            print_progress_bar(
                0, len(self._va_plugins_to_compile), prefix="Compiling Xyce plugins", length=50
            )
            processes = []
            for path_plugin, vafile in self._va_plugins_to_compile:
                # write to location
                vafile.rename_root(path_plugin.with_suffix(".va").name)
                # does it exist now ? (can happen if two different duts are used...)
                if path_plugin.is_file():
                    continue
                vafile.write_files(path_plugin.parent)

                # call "buildxyceplugin.sh"
                log_file = open(path_plugin.parent / (path_plugin.stem + "_compile.log"), "w")
                processes.append(
                    subprocess.Popen(
                        [
                            self.build_xyce_plugin_command,
                            "-o",
                            path_plugin.stem,
                            vafile.root,
                            ".",
                        ],
                        shell=False,
                        cwd=path_plugin.parent,
                        stderr=subprocess.STDOUT,
                        stdout=log_file,
                    )
                )

            time_start = time.time()
            time_out = True
            while time.time() - time_start < 60:
                finished = 0
                for i_p, process in enumerate(processes[:]):
                    returncode = process.poll()
                    if returncode is None:
                        time.sleep(0.1)
                    else:
                        if (
                            returncode == 0
                            or returncode != 0
                            # and returncode != 134
                            # and returncode != 139
                            # and returncode != 1
                        ):
                            print_progress_bar(
                                finished,
                                len(self._va_plugins_to_compile),
                                prefix="Compiling Xyce plugins",
                                length=50,
                            )
                            logpath = Path(process.args[3][:-3] + "_compile.log")
                            del processes[i_p]
                            finished += 1
                        else:
                            logpath = Path(process.args[3][:-3] + "_compile.log")
                            raise OSError(
                                "The plugin was not compiled successfull! Check the log in ",
                                str(logpath),
                                "content:",
                                logpath.with_name("buildxyceplugin.log").read_text(),
                            )

                if not processes:
                    time_out = False
                    break

            if time_out:
                raise OSError("The plugin was not compiled in time!")

            print("\n")  # new line after the progress bar

        self._va_plugins_to_compile = []

        # temperature and AC sweepdefs
        i_swd_temperature = None
        swd_ac = None
        for i_swd, swd in enumerate(tmp_sweep.sweepdef):
            if swd.var_name == specifiers.TEMPERATURE:
                if i_swd_temperature is None:
                    i_swd_temperature = i_swd
                else:
                    raise OSError("For Xyce only one temperature sweep is possible in one file!")
            elif swd.var_name == specifiers.FREQUENCY:
                if swd_ac is None:
                    swd_ac = tmp_sweep.sweepdef.pop(i_swd)
                else:
                    raise OSError("For Xyce only one AC simulation is possible in one file!")

        # get the header and add the simulation controllers
        str_netlist = self.inp_header + "\n"

        # noise ?
        if "noise" in tmp_sweep.outputdef:
            tmp_sweep.outputdef = [out_var for out_var in sweep.outputdef if out_var != "noise"]
            noise = True
            raise NotImplementedError(
                "For Xyce noise simulations are not implemented. Either implement it, pay Mario or use ADS."
            )
        else:
            noise = False

        # simulations
        # add a table with all voltages, frequencies and even temperature Oo
        outputdef = copy.deepcopy(tmp_sweep.outputdef)
        tmp_sweep.outputdef = []

        # ac forward and backward ?
        str_ac_switch = ""
        if swd_ac is not None:
            if "ac_switch" in str_netlist:
                tmp_sweep.sweepdef.append(
                    SweepDef(SpecifierStr("ac_switch"), "LIST", sweep_order=1000, value_def=[0, 1])
                )  # just use a very high sweep order to make sure it is the last one...
                str_ac_switch = " ac_switch"

        # remove columns which are not in the netlist
        tmp_sweep = tmp_sweep.set_values()
        df_def = tmp_sweep.create_df()
        for col in copy.deepcopy(df_def.columns):
            if col == specifiers.TEMPERATURE:
                df_def[col] = df_def[col] - constants.P_CELSIUS0  # convert to celsius
            elif str(col) not in str_netlist:
                df_def.drop(columns=col, inplace=True)

        str_netlist += "* Table of operation points and temperatures: \n"
        simulations = df_def.to_string(header=True, index=False)
        lines_simulations = simulations.split("\n")
        for i_line, line in enumerate(lines_simulations):
            lines_simulations[i_line] = "+ " + line
        simulations = "\n".join(lines_simulations)

        str_netlist += ".DATA TAB_SIMS\n" + simulations + "\n.ENDDATA\n\n"

        # DC operation point
        str_netlist += "* step through ops simulation\n"
        str_netlist += ".STEP DATA=TAB_SIMS\n"
        str_netlist += "* DC OP simulation\n"
        str_netlist += ".OP \n"
        # print the DC outputs
        str_netlist += "* DC OP output definition\n"
        str_netlist += (
            ".PRINT AC_IC FORMAT=CSV FILE=DC.csv TEMP V(*) I(*)" + str_ac_switch + " \n"
        )  # TODO add outputdef ?!?

        # AC simulation
        if swd_ac is not None:
            str_netlist += "* AC simulation\n"
            str_netlist += ".AC DATA=TAB_FREQUENCIES \n"
            str_netlist += ".DATA TAB_FREQUENCIES \n"
            str_netlist += "+ FREQ \n"
            str_netlist += "+ " + " ".join(f"{val:.6g}" for val in swd_ac.values) + " \n"
            str_netlist += ".ENDDATA \n"
            str_netlist += "* AC output definition\n"
            str_netlist += ".PRINT AC FORMAT=CSV FILE=AC.csv V(*) I(*) \n"  # TODO add outputdef ?!?
        else:
            raise NotImplementedError("Untested")

        # end simulation and log it
        str_netlist += "\n.END\n"

        logging.info("Added bias sweep to input header.")
        logging.debug("\n%s", str_netlist)

        return str_netlist

    def validate_simulation_successful(self, sweep):
        """Checks if the simulation of the given sweep was successful.

        Parameters
        ----------
        sweep  :  :class:`DMT.core.sweep.Sweep`
            Sweep that has been simulated.

        Raises
        ------
        NotImplementedError
            If the Dut is not a simulatable dut.
        SimulationUnsuccessful
            If the simulation output is not valid.
        FileNotFoundError
            If the sim log file does not exist.
        """
        # get sweep folder
        sim_folder = self.get_sim_folder(sweep)

        # check log file
        sim_log = sim_folder / "sim.log"

        log_content = sim_log.read_text()

        if "MSG_ERROR" in log_content:
            raise SimulationUnsuccessful(
                "Xyce simulation in the folder " + str(sim_folder) + " failed."
            )

        if "MSG_FATAL" in log_content:
            raise SimulationUnsuccessful(
                "Xyce simulation in the folder " + str(sim_folder) + " failed."
            )

    def import_output_data(self, sweep, delete_sim_results=False, key=None):
        """Read the output files that have been produced while simulating sweep and attach them to self.db.

        Parameters
        ----------
        sweep  :  :class:`DMT.core.sweep.Sweep`
            Sweep that has been simulated for the desired output files.
        delete_sim_results : {False, True}, optional
            If True, the simulation folder is deleted after reading.

        Raises
        ------
        NotImplementedError
            If the Dut is not a simulatable dut.
        IOError
            If the given sweep can not be read.
        """
        if isinstance(sweep, Sweep):
            # get sweep folder
            sim_folder = self.get_sim_folder(sweep)

            # find .raw file
            filepaths = list(sim_folder.glob("*.csv"))
            key = self.join_key(self.get_sweep_key(sweep), "iv")
        else:
            filepaths = [
                Path(sweep),
            ]
            sim_folder = filepaths[0].parent
            if key is None:
                warnings.warn("DMT->DutXyce->import_output_data: Added a df with the key None!")

        df = self._import_xyce_results(*filepaths)

        self.data[key] = df

        if isinstance(sweep, Sweep):
            logging.info(
                "Read the Xyce simulation output data of the sweep %s. \nThe simulation folder is %s",
                sweep.name,
                sim_folder,
            )

            sim_log = sim_folder / "sim.log"
            log_content = sim_log.read_text()
            search_obj_time = re.search(
                r"Total Simulation Solvers Run Time: ([\d*?\.\d*?]) seconds",
                log_content,
                flags=re.IGNORECASE | re.DOTALL,
            )
            if search_obj_time:
                logging.info("Simulation time: %s.", search_obj_time.group(1))

            if delete_sim_results:
                self.delete_sim_results(sweep)
        else:
            logging.info(
                "Read the Xyce simulation output data of the from the file %s.", filepaths[0]
            )

    def _import_xyce_results(self, *filepaths):
        """Reads the filepaths (must be CSVs), converts them and joins them together to one DMT.DataFrame

        Parameters
        ----------
        filenames : os.Pathlike
            Paths to the *.csv files.
        """
        col_ac_switch = SpecifierStr("AC_SWITCH")
        df_main = None

        for filepath in filepaths:
            df = pd.read_csv(filepath)
            df.__class__ = DataFrame  # cast it

            # convert column names to specifiers
            to_drop = []  # drop after iteration!
            columns = {}
            for i_col, col in enumerate(df.columns):
                if col.startswith("V("):  # is this a potential
                    node = col[4:-1]  # example: V(N_E_FORCED) -> E_FORCED
                    if "FORCED" in node:
                        col = (
                            specifiers.VOLTAGE + node[0:-7] + sub_specifiers.FORCED
                        )  # example: E_FORCED -> E
                    else:
                        col = specifiers.VOLTAGE + node
                elif col.startswith("I("):  # current
                    device = col[2:-1]
                    if device.startswith("R"):
                        col = specifiers.CURRENT + device[3:]
                    else:
                        col = specifiers.CURRENT + device
                elif col.startswith("Re("):  # real part of something
                    # Y parameters and also dV/dV (what to do with those ?!?)
                    if col[3] == "V":
                        node = col[7:-2]  # example: Re(V(N_B_FORCED)) -> B_FORCED
                        if "FORCED" in node:
                            col = (
                                specifiers.VOLTAGE
                                + node[0:-7]
                                + sub_specifiers.FORCED
                                + sub_specifiers.AC
                                + sub_specifiers.REAL
                            )  # example: E_FORCED -> E
                        else:
                            col = (
                                specifiers.VOLTAGE + node + sub_specifiers.AC + sub_specifiers.REAL
                            )
                    elif col[3] == "I":
                        device = col[5:-2]
                        if device.startswith("R"):
                            col = specifiers.SS_PARA_Y + device[3:] + sub_specifiers.REAL
                        else:
                            col = specifiers.SS_PARA_Y + device + sub_specifiers.REAL
                    else:
                        raise NotImplementedError("What is this??")
                elif col.startswith("Im("):  # imaginary part of something
                    # Y parameters and also dV/dV (what to do with those ?!?)
                    if col[3] == "V":
                        node = col[7:-2]  # example: Re(V(N_B_FORCED)) -> B_FORCED
                        if "FORCED" in node:
                            col = (
                                specifiers.VOLTAGE
                                + node[0:-7]
                                + sub_specifiers.FORCED
                                + sub_specifiers.AC
                                + sub_specifiers.IMAG
                            )  # example: E_FORCED -> E
                        else:
                            col = (
                                specifiers.VOLTAGE + node + sub_specifiers.AC + sub_specifiers.IMAG
                            )
                    elif col[3] == "I":
                        device = col[5:-2]
                        if device.startswith("R"):
                            col = specifiers.SS_PARA_Y + device[3:] + sub_specifiers.IMAG
                        else:
                            col = specifiers.SS_PARA_Y + device + sub_specifiers.IMAG
                    else:
                        raise NotImplementedError("What is this??")
                elif col == specifiers.TIME:
                    if np.any(df[col]):
                        col = specifiers.TIME
                    else:  # it is 0 for all rows in a AC_OP STEP simulation
                        to_drop.append(col)
                elif col == specifiers.FREQUENCY:
                    col = specifiers.FREQUENCY
                elif col == specifiers.TEMPERATURE:
                    col = specifiers.TEMPERATURE
                elif col == col_ac_switch:
                    col = col_ac_switch
                else:
                    raise NotImplementedError("Implement the correct specifier!")

                columns[df.columns[i_col]] = col  # pylint: disable=unsubscriptable-object

            df.drop(columns=to_drop, inplace=True)

            df = df.rename(columns=columns)

            # join the real and imaginary to values
            for i_col, col in enumerate(df.columns):
                try:
                    if sub_specifiers.REAL.sub_specifiers[0] in col.sub_specifiers:
                        col_complex = SpecifierStr(col.specifier) + col.nodes
                        for sub_spec in col.sub_specifiers:
                            if sub_spec != sub_specifiers.REAL.sub_specifiers[0]:
                                col_complex += sub_spec
                        col_imag = col_complex + sub_specifiers.IMAG
                        df[col_complex] = df[col] + 1j * df[col_imag]
                        df.drop(columns=[col, col_imag], inplace=True)
                except AttributeError:
                    # should not happen since all columns should be specifiers right now
                    raise  # but if it happens just comment the exception (if no time to fix!)

            # add the columns to df_main
            if df_main is None:
                df_main = df
            else:
                df_ac = None
                df_dc = None
                if (
                    specifiers.FREQUENCY in df.columns
                    and specifiers.FREQUENCY not in df_main.columns
                ):
                    df_dc = df_main
                    df_ac = df
                elif (
                    specifiers.FREQUENCY not in df.columns
                    and specifiers.FREQUENCY in df_main.columns
                ):
                    df_dc = df
                    df_ac = df_main
                else:
                    raise IOError("Join with main depending on content :/")

                # repeat each line in df_dc by number of frequencies
                len_frequencies = np.unique(df_ac[specifiers.FREQUENCY].to_numpy()).size
                df_main = df_dc.repeat_rows(len_frequencies)
                df_main = df_main.join(df_ac)

        if col_ac_switch in df_main.columns:

            # get all AC columns, except frequency and ac_switch
            cols_ac_old = [
                col for col in df_ac.columns if col not in [specifiers.FREQUENCY, col_ac_switch]
            ]

            #  the AC columns to account for forward and backward simulation
            if self.dut_type.is_subtype(DutType.flag_bjt):
                ports = ["B", "C"]
            elif self.dut_type.is_subtype(DutType.flag_mos):
                ports = ["G", "D"]
            else:
                raise NotImplementedError("DutType not known!")
            cols_ac_forward = []
            cols_ac_backward = []
            for col_old in cols_ac_old:
                cols_ac_forward.append(
                    SpecifierStr(col_old.specifier)
                    + [*col_old.nodes, ports[0]]
                    + col_old.sub_specifiers
                )
                cols_ac_backward.append(
                    SpecifierStr(col_old.specifier)
                    + [*col_old.nodes, ports[1]]
                    + col_old.sub_specifiers
                )
            # get the DC columns
            cols_dc = [col for col in df_dc.columns if col not in [col_ac_switch]]

            # find the row pairs with same forced dc, temp and frequency
            cols_filter = [
                col
                for col in cols_dc
                if sub_specifiers.FORCED.sub_specifiers[0] in col.sub_specifiers
            ]
            cols_filter.append(specifiers.TEMPERATURE)
            cols_filter.append(specifiers.FREQUENCY)

            row_pairs = set()
            for i_row, row in df_main.iterrows():
                sub_data = df_main
                for col_filter in cols_filter:
                    sub_data = sub_data[np.isclose(sub_data[col_filter], row[col_filter])]

                # now there should be exactly 2 rows left
                if len(sub_data) == 2:
                    row_pairs |= {(*sub_data.index.to_list(),)}
                else:
                    raise IOError("More than 2 rows for one 'AC_SWITCH' ?!?")

            # make a new dataframe
            df_main_new = DataFrame(
                columns=cols_dc + [specifiers.FREQUENCY] + cols_ac_forward + cols_ac_backward,
                dtype=np.complex64,
            )
            # join the rows, drop the old rows and save the new row in the dataframe
            for i_row_1, i_row_2 in row_pairs:
                series_dc = df_main.loc[i_row_1, cols_dc]

                if df_main.at[i_row_1, col_ac_switch]:
                    series_ac_forward = df_main.loc[i_row_2, cols_ac_old]
                    series_ac_backward = df_main.loc[i_row_1, cols_ac_old]
                elif df_main.at[i_row_2, col_ac_switch]:
                    series_ac_forward = df_main.loc[i_row_1, cols_ac_old]
                    series_ac_backward = df_main.loc[i_row_2, cols_ac_old]
                else:
                    raise IOError("Did not find forward and backward simulation")

                series_ac_forward.rename(dict(zip(cols_ac_old, cols_ac_forward)), inplace=True)
                series_ac_backward.rename(dict(zip(cols_ac_old, cols_ac_backward)), inplace=True)

                # do not forget the frequency
                series_ac_forward[specifiers.FREQUENCY] = df_main.loc[i_row_1, specifiers.FREQUENCY]

                # add the new row to the new dataframe
                df_main_new = pd.concat(
                    [
                        df_main_new,
                        pd.concat([series_dc, series_ac_forward, series_ac_backward]).to_frame().T,
                    ],
                    axis="index",
                    ignore_index=True,
                )

            df_main = df_main_new.sort_values(by=cols_filter)

        # unit converts!
        if specifiers.TEMPERATURE in df_main.columns:
            df_main[specifiers.TEMPERATURE] = (
                df_main[specifiers.TEMPERATURE] + constants.P_CELSIUS0
            )  # in DMT everything in Kelvin -.-

        return df_main



""" Converter routine to create a sweep definition from a given DataFrame
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
from __future__ import annotations
from typing import Type, Optional
import warnings
import numpy as np

from DMT.core import Sweep, specifiers, sub_specifiers, DataFrame, SweepDef


def df_to_sweep(
    df_to_convert: DataFrame,
    temperature: float = 300.0,
    name: Optional[str] = None,
    from_forced: bool = True,
    SweepDefClass: Type = SweepDef,
    decimals_potentials: int = 3,
):
    """Create a sweepdefinition which simulates the same forced values as the given measurement data frame.

    For the moment only voltage sweeps...

    Parameters
    ----------
    df_to_convert : :class:`~DMT.core.data_frame.DataFrame`
        DataFrame with voltages which define the sweep.
    temperature : float, optional
        Temperature of the measurement in Kelvin, defaults to 300.0
    name : str, optional
        Name of the sweep to create. If None, automatic generation of a name is tried.
    from_forced : {True, False}, optional
        If True, the forced voltages are used to create the sweepdef.
    SweepDefClass : :class:`~DMT.core.sweep.SweepDef`
        SweepDef class for the sweep to create from
    decimals_potentials : int
        Round the potentials to this x decimals, defaults to 3

    Returns
    -------
    :class:`~DMT.core.sweep.Sweep`
    """
    warnings.warn(
        "df_to_sweep is deprecated and will be removed in the next major release.\nUse Sweep.get_sweep in future.\n",
        category=DeprecationWarning,
    )

    vars_other = {
        specifiers.TEMPERATURE: temperature
    }  # temperature is usually in the dut.data key and not in the df. Additionally it is always constant ?!?
    def_output = []
    potentials_forced = []  # list of forced voltages
    for col in df_to_convert.columns:
        try:
            if col.specifier not in [
                specifiers.VOLTAGE,
                specifiers.FREQUENCY,
            ]:  # all columns except voltages and frequency are outputs
                def_output.append(col)
            if (
                (col.specifier == specifiers.VOLTAGE)
                and len(col.nodes) == 1
                and (not from_forced or (sub_specifiers.FORCED in col))
            ):  # forced voltages are needed to create sweepdef
                potentials_forced.append(col)
        except AttributeError:  # can only analyze specifiers
            def_output.append(col)

    if not potentials_forced:
        raise IOError(
            "DMT->df_to_sweep: No forced potentials in the columns. Make sure all forced potentials in the columns are SpecifierStr!"
        )

    # find one potential with only one value and set it as reference potential!
    potential_common = ""
    for potential in sorted(
        potentials_forced, key=lambda x: x.nodes[0]
    ):  # sort to make sure everytime the same is found.
        if len(np.unique(np.round(df_to_convert[potential].to_numpy(), decimals_potentials))) == 1:
            potential_common = potential  # use this one! Even if more than one would exist.
            break

    if not potential_common:  # no potential found
        # did not find anything :(
        raise IOError(
            "DMT -> df_to_sweep: The user supplied voltages do not define unambiguous sweep conditions for the transistor. No common reference potential was found."
        )

    def_sweep = [
        {"var_name": potential_common, "sweep_order": 0, "sweep_type": "CON", "value_def": [0]},
    ]  # 0 because the other Sweepdefs are added as voltages.
    sweep_order = 1

    # now we need all sweep potentials with regard to one reference potential or other synced sweeps.
    potentials_sweep = [
        potential for potential in set(potentials_forced) if potential != potential_common
    ]

    # find order of sweeps including V_BC sweeps... (innermost, ..., outermost)
    count_ops = {}
    for potential in sorted(potentials_sweep, key=lambda x: x.nodes[0]):
        # against all other:
        for potential_other in sorted(potentials_forced, key=lambda x: x.nodes[0]):
            if potential_other == potential:
                continue  # not against self

            if from_forced:
                if potential_other + potential.nodes[0] + sub_specifiers.FORCED in count_ops:
                    continue  # not V_BC and V_CB, only one of those..

                voltage = potential + potential_other.nodes[0] + sub_specifiers.FORCED
            else:
                if potential_other + potential.nodes[0] in count_ops:
                    continue  # not V_BC and V_CB, only one of those..

                voltage = potential + potential_other.nodes[0]

            df_to_convert.ensure_specifier_column(voltage)
            values = np.unique(np.round(df_to_convert[voltage].to_numpy(), decimals_potentials))
            count_ops[voltage] = len(values)

    for i_voltage, (voltage, _count) in enumerate(
        sorted(count_ops.items(), key=lambda kv: kv[1])[:-1]
    ):  # last of the sorted list is the one we don't need
        df_to_convert.ensure_specifier_column(voltage)
        values = np.unique(np.round(df_to_convert[voltage].to_numpy(), decimals_potentials))

        if voltage.nodes[1] == potential_common.nodes[0]:  # convert to potential if possible
            if from_forced:
                potential = specifiers.VOLTAGE + voltage.nodes[0] + sub_specifiers.FORCED
            else:
                potential = specifiers.VOLTAGE + voltage.nodes[0]

            def_sweep.append(
                {
                    "var_name": potential,
                    "sweep_order": sweep_order,
                    "sweep_type": "LIST",
                    "value_def": values,
                }
            )
        else:  # sweep a voltage is more difficult
            # nodes in all other voltages:
            other_nodes = ""
            for voltage_other, _count in sorted(count_ops.items(), key=lambda kv: kv[1])[
                i_voltage + 1 : -1
            ]:
                other_nodes += "".join(voltage_other.nodes)

            potential = specifiers.VOLTAGE + "C"  # per default a BC sweep
            potential_master = specifiers.VOLTAGE + "B"
            for node_master in voltage.nodes:
                if node_master in other_nodes:
                    if from_forced:
                        potential_master = specifiers.VOLTAGE + node_master + sub_specifiers.FORCED
                        node_other = next(node for node in voltage.nodes if node != node_master)
                        potential = specifiers.VOLTAGE + node_other + sub_specifiers.FORCED
                    else:
                        potential_master = specifiers.VOLTAGE + node_master
                        node_other = next(node for node in voltage.nodes if node != node_master)
                        potential = specifiers.VOLTAGE + node_other

            voltage_offset = potential + potential_master.nodes[0]

            if voltage != voltage_offset:  # order of nodes reversed
                values = np.round(-values, decimals_potentials)  # invert voltage

            def_sweep.append(
                {
                    "var_name": voltage_offset,
                    "sweep_order": sweep_order,
                    "sweep_type": "LIST",
                    "value_def": values,
                }
            )

            # find sweep_order of potential_master
            sweep_order_master = 0
            for i_voltage_master, (voltage_master, _count) in enumerate(
                sorted(count_ops.items(), key=lambda kv: kv[1])[:-1]
            ):
                if (
                    voltage_master.nodes[1] == potential_common.nodes[0]
                ):  # convert to potential if possible
                    if voltage_master.nodes[0] == potential_master.nodes[0]:
                        sweep_order_master = i_voltage_master + 1  # plus 1 bc 1st swd is root
                        break

            def_sweep.append(
                {
                    "var_name": potential,
                    "sweep_order": sweep_order_master,
                    "sweep_type": "SYNC",
                    "master": potential_master,
                    "offset": voltage_offset,
                }
            )

        sweep_order += 1

    ### grab definition for the frequency, if possible
    if specifiers.FREQUENCY in df_to_convert.columns:
        freq = np.unique(
            np.round(df_to_convert[specifiers.FREQUENCY].to_numpy(), decimals_potentials)
        )
        def_sweep.append(
            {
                "var_name": specifiers.FREQUENCY,
                "sweep_order": sweep_order,
                "sweep_type": "LIST",
                "value_def": freq,
            },
        )
        is_ac = True
    else:
        is_ac = False

    ### Name
    if name is None:
        # generate a (good) name
        if is_ac:
            name = "AC"
        else:
            name = "DC"

    return Sweep(
        name,
        sweepdef=def_sweep,
        othervar=vars_other,
        outputdef=def_output,
        SweepDefClass=SweepDefClass,
    )



""" DMT description of a circuit.

Must be used to describe a circuit and then passed to a circuit simulator dut.

Later on this can be extended to allow (pseudo-)simulations directly inside DMT.

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
from __future__ import annotations
import warnings
from typing import Callable, Iterable, Optional, Union, List, Tuple
from DMT.core import MCard, McParameterCollection


RESISTANCE = "R"
""" Indicates a resistance in a circuit. """
CAPACITANCE = "C"
""" Indicates a capacitance in a circuit. """
INDUCTANCE = "L"
""" Indicates a inductance in a circuit. """
CURRENT = "I_Source"
""" Indicates a current source in a circuit. """
VOLTAGE = "V_Source"
""" Indicates a voltage source in a circuit. """
SHORT = "Short"
""" Indicates a voltage source in a circuit. """
HICUML2_HBT = "BHT"
""" Indicates a bipolar transistor modeled with HICUM/L2 in a circuit. """
HICUML0_HBT = "BHT0"
""" Indicates a bipolar transistor modeled with HICUM/L0 in a circuit. """
SGP_BJT = "BJT"
""" Indicates a bipolar transistor modeled with SGP in a circuit. """
DIODE = "diode"
""" Indicates a diode in a circuit. """


class CircuitElement(object):
    """Class that is used to describe netlist elements such as resistors, capacitors or inductors.

    The possible CircuitElements define the DMT netlist format. Circuit simulators need to convert the DMT netlist format to their respective one.
    This class has a special emphasis on good error messages and typesetting.

    Possible element types:

    * :py:const:`RESISTANCE`
    * :py:const:`CAPACITANCE`Sequence
    * :py:const:`SHORT`
    * va_module -> could be a transistor

    Parameters
    ----------
    element_type
        Element type, for example: 'R' for resistor. possible values: 'V_Source', 'I_Source', 'R', 'C', 'L', 'Short' ,'va_module'
    name
        Element name, for example: 'R1' for resistor 1. Names should be unique within their netlist.
    contact_nodes
        Contact nodes of the element, for example: ('n__1', 'n__2')
    parameters
        Parameters of the element, for example: [('R', '1k')]

    Attributes
    ----------
    element_type
        Element type, for example: 'R' for resistor
    name
        Element name, for example: 'R1' for resistor 1
    contact_nodes
        Contact nodes of the element, for example: ('n__1', 'n__2')
    parameters
        Parameters of the element, for example: [('R', '1k')]
    """

    possible_types = [
        VOLTAGE,
        CURRENT,
        RESISTANCE,
        CAPACITANCE,
        INDUCTANCE,
        SHORT,
        HICUML2_HBT,
        HICUML0_HBT,
        SGP_BJT,
        DIODE,
        "va_module",
        "pdk",
        '"hbt_n1s"',
        "n1s_sgp",
        "hbt_n1s",
        "hbt_n2s",
        '"hbt_n2s"',
        "hbt_n3s",
        '"npn_hicum"',
        '"npn_hs"',
        "npn_hs",
        "TSC250_Models_lib_TSC_250nm_Agilent_v1p0_schematic",
        '"TSC250_Models_lib_TSC_250nm_Agilent_v1p0_schematic"',
    ]

    def __init__(
        self,
        element_type: str,
        name: str,
        contact_nodes: Iterable[str],
        parameters: Optional[Union[List[Tuple[str, str]], MCard, McParameterCollection]] = None,
        method: Optional[Callable] = None,
    ):
        if isinstance(parameters, MCard) or isinstance(parameters, McParameterCollection):
            CircuitElement.possible_types.append(parameters.default_module_name)  # type: ignore

        if isinstance(element_type, str):
            if element_type in self.possible_types:
                self.element_type = element_type

            else:
                raise IOError(
                    "DMT -> Circuit: Element Type "
                    + str(element_type)
                    + " is unknown. \n Possible types: "
                    + str(CircuitElement.possible_types)
                )

        else:
            raise TypeError("DMT -> element_type: element_type needs to be a string.")

        if isinstance(name, str):
            self.name = name
        else:
            raise TypeError(
                "The element name has to be a string! Given was "
                + str(name)
                + " of type "
                + type(name)
            )

        if isinstance(contact_nodes, (tuple, list)):
            for i_node, node in enumerate(contact_nodes):
                if not isinstance(node, str):
                    raise TypeError(
                        "The element contact nodes have to be a tuple of strings! Given was a tuple of different elements. The node "
                        + str(i_node)
                        + " had the type "
                        + type(node)
                    )

            self.contact_nodes = contact_nodes

        else:
            raise TypeError(
                "The element contact nodes have to be a tuple of strings! Given was "
                + str(contact_nodes)  # type: ignore
                + " of type "
                + type(contact_nodes)
            )

        if parameters is None:
            pass

        elif isinstance(parameters, list):
            for i_parameter, parameter in enumerate(parameters):
                if isinstance(parameter, tuple):
                    for parameter_part in parameter:
                        if not isinstance(parameter_part, str):
                            raise TypeError(
                                "The parameters have to be a list of tuple of strings! Given was a list of tuples with at least one element of "
                                + type(parameter_part)
                                + " in the tuple "
                                + str(i_parameter)
                            )
                else:
                    raise TypeError(
                        "The parameters have to be a list of tuple of strings! Given was a list with at least one element of "
                        + type(parameter)
                        + " at position "
                        + str(i_parameter)
                    )

        elif isinstance(parameters, MCard) or isinstance(parameters, McParameterCollection):
            # Allow model cards!
            pass

        else:
            raise TypeError(
                "The parameters have to be a list of tuple of strings or a modelcard! Given was a "
                + type(parameters)
            )

        self.parameters = parameters

        if method is not None:
            warnings.warn(
                "method will be deprecated in future DMT releases. This feature is not needed anymore since VAE is mature.",
                category=DeprecationWarning,
            )
        self.method = method

    def __repr__(self):
        str_nodes = ",".join(self.contact_nodes)
        return f"DMT.CircuitElement:{self.name:s} model {self.element_type:s} nodes {str_nodes:s}"


class Circuit(object):
    """Circuit description as a list of :class:`CircuitElement`

    Parameters
    ----------
    circuit_elements
        Either directly the netlist elements as a list of CircuitElements or strings (for equations)

    Attributes
    ----------
    netlist
        Either directly the netlist elements as a list of CircuitElements or strings (for equations)
    """

    def __init__(self, circuit_elements: List[Union[str, CircuitElement]]):
        if isinstance(circuit_elements, str):
            raise NotImplementedError(
                "The default circuit has been moved into the corresponding module. Access the circuit from there!"
            )

        for i_element, element in enumerate(circuit_elements):
            if not isinstance(element, (CircuitElement, str)):
                raise TypeError(
                    "The netlist has to be a list of CircuitElement or str! At position "
                    + str(i_element)
                    + " is a entry of type "
                    + type(element)
                )

        self.netlist = circuit_elements



# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
name = "xyce"

from .dut_xyce import DutXyce

# grab the verilog-AMS codes from
# https://github.com/Xyce/Xyce/tree/master/utils/ADMS/examples



""" Used to flag the dut type for each dut view

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
from enum import Flag, auto, unique


class DutTypeInt(object):
    """Class for DutType flags. Adds the nodes attribute to an integer value. This allows direct assignment of nodes to a DutType.

    Parameters
    ----------
    value : int
    node  : [str]
        List of nodes.
    """

    def __init__(self, value, nodes=None, string=None):
        self.nodes = []
        self.string = string
        try:
            # is value a DutTypeInt?
            self.value = value.value
            self.nodes = value.nodes
        except AttributeError:
            # or is it a integer:
            self.value = value

        # if nodes is self.nodes is still none or nodes was given as a parameter overwrite it!
        if nodes is not None:
            self.nodes = nodes

    def get_nodes(self):
        """Return the nodes that are typically found in this Dut_type. For convenience.

        Repeated here just to get rid of the pylint error. The real method is below in the DutType-flag

        Returns
        -------
        nodes  :  list of strings
            List of strings
        """
        return self.value.nodes

    def get_string(self):
        """Return the string that describes this Dut_type.

        Returns
        -------
        string  :  string
            the string that describes this object
        """
        return self.string

    def __and__(self, other):
        try:
            return DutTypeInt(self.value & other.value, nodes=self.nodes)
        except AttributeError:
            return DutTypeInt(self.value & other, nodes=self.nodes)

    def __rand__(self, other):
        return self.__and__(other)

    def __xor__(self, other):
        try:
            return DutTypeInt(self.value ^ other.value, nodes=self.nodes)
        except AttributeError:
            return DutTypeInt(self.value ^ other, nodes=self.nodes)

    def __rxor__(self, other):
        return self.__xor__(other)

    def __or__(self, other):
        try:
            return DutTypeInt(self.value | other.value, nodes=self.nodes)
        except AttributeError:
            return DutTypeInt(self.value | other, nodes=self.nodes)

    def __ror__(self, other):
        return self.__or__(other)

    def __eq__(self, other):
        try:
            return self.value == other.value
        except AttributeError:
            return self.value == other

    def __invert__(self):
        return DutTypeInt(~self.value, nodes=self.nodes)

    def __bool__(self):
        return bool(self.value)

    def __str__(self):
        return self.string

    def __int__(self):
        return self.value

    def is_subtype(self, other):
        # only for pylint!
        pass

    def __lt__(self, other):
        try:
            return DutTypeInt(self.value < other.value, nodes=self.nodes)
        except AttributeError:
            return DutTypeInt(self.value < other, nodes=self.nodes)

    def __le__(self, other):
        try:
            return DutTypeInt(self.value <= other.value, nodes=self.nodes)
        except AttributeError:
            return DutTypeInt(self.value <= other, nodes=self.nodes)

    def __gt__(self, other):
        try:
            return DutTypeInt(self.value > other.value, nodes=self.nodes)
        except AttributeError:
            return DutTypeInt(self.value > other, nodes=self.nodes)

    def __ge__(self, other):
        try:
            return DutTypeInt(self.value >= other.value, nodes=self.nodes)
        except AttributeError:
            return DutTypeInt(self.value >= other, nodes=self.nodes)

    def __hash__(self):
        return hash((self.value, tuple(self.nodes), self.string))

    def bit_length(self):
        return self.value.bit_length()


@unique  # do not allow same values for different types
class DutType(Flag):
    """
    Flags which represents all common devices that might need to be handled by DMT

    Methods
    -------
    get_nodes(filename)
        Returns the names of the nodes typicall found in the specified Dut_type
    """

    # # get the main non mixed flags, there are pure integers because auto() does not return a value but it is a generator i don't really know how to use:
    _flag_subtype_1 = auto()
    _flag_subtype_2 = auto()
    _flag_subtype_3 = auto()
    _flag_subtype_4 = auto()
    device = auto()
    bulk = auto()
    meas_struct = auto()
    deemb_struct = auto()
    flag_transistor = auto()
    flag_bjt = auto()
    flag_bjt_deemb = auto()
    flag_mos = auto()
    flag_mos_deemb = auto()
    flag_diode = auto()
    flag_cap = auto()
    flag_res = auto()
    flag_tetrode = auto()
    flag_tlm = auto()
    flag_deem = auto()
    flag_vdp = auto()

    flag_npn = auto()
    flag_pnp = auto()
    flag_n_mos = auto()
    flag_p_mos = auto()

    flag_open = auto()
    flag_short = auto()

    dummy = DutTypeInt(0)  # dummy is nothing! Use DutTypeInt to allow get_nodes

    # now the mixed flags, these can be DutTypeInts as the numbers are already given:
    transistor = DutTypeInt(device | flag_transistor)
    bjt = DutTypeInt(transistor | flag_bjt, nodes=["B", "C", "E", "S"], string="bjt")
    mos = DutTypeInt(transistor | flag_mos, nodes=["G", "D", "S", "B"], string="mos")
    deem_bjt = DutTypeInt(
        deemb_struct | flag_bjt_deemb, nodes=["B", "C", "E", "S"], string="bjt_deemb"
    )  # because of node names :(
    deem_mos = DutTypeInt(
        deemb_struct | flag_mos_deemb, nodes=["G", "D", "S", "B"], string="mos_deemb"
    )  # because of node names :(

    npn = DutTypeInt(bjt | flag_npn, string="npn")  # nodes are inherited from bjt
    pnp = DutTypeInt(bjt | flag_pnp, string="pnp")
    n_mos = DutTypeInt(mos | flag_n_mos, string="nmos")
    p_mos = DutTypeInt(mos | flag_p_mos, string="pmos")

    diode = DutTypeInt(device | flag_diode, nodes=["C", "A"], string="diode")
    pn_diode = DutTypeInt(diode | _flag_subtype_1, string="pn-diode")
    pin_diode = DutTypeInt(diode | _flag_subtype_2, string="pin-diode")
    cap = DutTypeInt(device | flag_cap, nodes=["C", "A"], string="capacitance")
    res = DutTypeInt(device | flag_res, nodes=["C", "A"], string="resistor")

    tlm = DutTypeInt(
        meas_struct | flag_tlm, nodes=["L", "M", "R"], string="tlm"
    )  # left, middle, right
    tlmb = DutTypeInt(tlm | _flag_subtype_1, string="tlm-base")
    tlmc = DutTypeInt(tlm | _flag_subtype_2, string="tlm-collector")
    tlmbc = DutTypeInt(tlm | _flag_subtype_3, string="tlm-base-collector")

    vdp = DutTypeInt(
        meas_struct | flag_vdp, nodes=["A", "B", "C", "D"], string="vdp"
    )  # four arbitrary contacts

    deem_open_bjt = DutTypeInt(deem_bjt | flag_open, string="open")
    deem_short_bjt = DutTypeInt(deem_bjt | flag_short, string="short")

    deem_open_mos = DutTypeInt(deem_mos | flag_open, string="open")
    deem_short_mos = DutTypeInt(deem_mos | flag_short, string="short")

    tetrode = DutTypeInt(
        meas_struct | flag_tetrode, nodes=["B1", "B2", "E", "C", "S"], string="tetrode"
    )

    cap_ac = DutTypeInt(
        cap | meas_struct, nodes=["L", "R", "G", "S"], string="capacitance-ac"
    )  # capacitance in GSG pads, each pad is one Capacitance, so S(1,1) and S(2,2) are wanted...

    def get_nodes(self):
        """
        Return the nodes that are typically found in this Dut_type. For convenience.

        Returns
        -------
        nodes  :  list of strings
            List of strings
        """
        try:
            return self.value.nodes
        except AttributeError:
            return ""
            # raise IOError('DMT -> Dut_type: Tried to get the nodes of a non-mixed device.')

    def get_string(self):
        """
        Return the string that describes this Dut_type.

        Returns
        -------
        nodes  :  string
            String that describes this Dut_type.
        """
        try:
            return self.string

        except AttributeError:
            return ""
            # raise IOError('DMT -> Dut_type: Tried to get the nodes of a non-mixed device.')

    def __str__(self):
        if self.value.string is not None:  # pylint: disable = no-member
            return self.value.string  # pylint: disable = no-member
        else:
            return str(self.value.__class__)

    def is_subtype(self, other):
        """Test if a device is a subtype of an other device/devicetype

        Ignores the flag_subtype!

        Parameters
        ----------
        other : int, DutTypeInt
        """
        try:
            val = int(self.value & other.value)
        except AttributeError:
            val = int(self.value & other)

        # remove subtype flag..
        n_subtype_1 = ~int(DutType._flag_subtype_1.value)
        n_subtype_2 = ~int(DutType._flag_subtype_2.value)
        n_subtype_3 = ~int(DutType._flag_subtype_3.value)
        n_subtype_4 = ~int(DutType._flag_subtype_4.value)

        val = (val & n_subtype_1) & (val & n_subtype_2) & (val & n_subtype_3) & (val & n_subtype_4)

        return bool(val)



""" DMT specific PyLaTeX classes and routines used for automatic documentations

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
from DMT.external.latex import build_tex

try:  # soft dependency: DMT can be used without pylatex...
    from pylatex import Section
    from pylatex import Package, NoEscape
    from pylatex.base_classes import CommandBase, Container
except ImportError:
    pass


class Tex(Section):
    """This is basically a Tex container as defined in Pylatex that can do nothing but contain Tex stuff."""

    def __init__(self, numbering=None, label=False, **kwargs):
        super().__init__(
            "dummy", numbering=None, label=label
        )  # dont need a title for a few tex commands

    def dumps(self):
        return self.dumps_content()


class Listing(Container):
    #: Default prefix to use with Marker
    marker_prefix = "listing"

    def __init__(
        self, content=None, file_path=None, language="TRADICA", caption=None, label=None, **kwargs
    ):
        super().__init__(**kwargs)

        self.language = language
        self.content = content
        self.file_path = file_path
        self.caption = caption
        self.label = label

    def dumps(self):
        if self.language == "TRADICA":
            str_language = (
                "backgroundcolor=\\color{lightgray!40},\n"
                + "breaklines=true, breakatwhitespace=true,\n"
                + "emph={[1]NAME, TRACE, OPTI, PHYS, TECH, DEVICE, OUTPUT, PGEN, RUN, END}, emphstyle={[1]\\bfseries},\n"
                + "emph={[2]UNIT_INP, PERMITT, BGP NARR, SAT_VELO, HOLE_MOBI, ELEC_MOBI, INFO_PROCESS, GEOM_DATA, ELEC_PAR, HICUM_PAR, MODEL_DEF, MISC_VAR, TRAN_CONF}, emphstyle={[2]\\color{Blue}},\n"
            )
        else:
            str_language = "language=" + self.language + ",\n"

        str_caption = "caption={:s},\n".format(self.caption) if self.caption is not None else ""
        str_label = "label={:s},\n".format(self.label) if self.label is not None else ""
        if self.file_path is None:
            return (
                "\\begin{lstlisting}[\n"
                + str_caption
                + str_label
                + str_language
                + "]\n"
                + self.content
                + "\\end{lstlisting}\n"
            )
        else:
            return (
                "\\lstinputlisting[\n"
                + str_caption
                + str_label
                + str_language
                + "]\n"
                + "{"
                + self.file_path
                + "}\n"
            )


class SubFile(Section):
    """This class implements a container to create SubFiles (thats a Latex package)"""

    def __init__(self, numbering=None, label=False, master=None, **kwargs):
        self.master = master
        if self.master is None:
            self.master = "documentation.tex"
        super().__init__(
            "dummy", numbering=None, label=label
        )  # dont need a title for a few tex commands

    def dumps(self):
        string = self.dumps_content()
        # now we add what is needed for subfiles
        string = (
            "\\documentclass["
            + self.master
            + "]{subfiles}\n"
            + "\\begin{document}\n"
            + string
            + "\n \\end{document} \n"
        )
        return string

    def generate_pdf(self, full_path_to_file, compiler=None):
        self.generate_tex(str(full_path_to_file))
        build_tex(full_path_to_file, additional_compiler=compiler)


class CommandRefRange(CommandBase):
    """This command can be used to reference multiple things like figures."""

    _latex_name = "crefrange"
    _default_escape = False
    packages = [Package("hyperref"), Package("cleveref")]


class CommandRef(CommandBase):
    """This command can be used to reference multiple things like figures."""

    _latex_name = "cref"
    _default_escape = False
    packages = [Package("hyperref"), Package("cleveref")]


class CommandInput(CommandBase):
    """This command can be used to input matplotlib2tikz .tex files."""

    _latex_name = "input"
    _default_escape = False
    packages = [
        Package("siunitx"),
        Package("tikz"),
        Package("fontspec"),
        Package("pgfplots"),
        Package("inputenc", options="utf8"),
    ]


class CommandLabel(CommandBase):
    """This command can be used to input matplotlib2tikz .tex files."""

    _latex_name = "label"
    _default_escape = False
    packages = []



# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import sys


def print_progress_bar(
    current=0,
    total=100,
    prefix="progress",
    suffix="completed",
    length=100,
    space=" ",
    fill="█",
    show_balance=False,
    completed=False,
):
    """Displays a progress bar in the terminal."""
    current = 100 if completed else current
    percentage = int((current * 100) / total)
    progress = int((percentage * length) / 100)
    bar = space * int(length - progress)
    progress = fill * progress

    balance = " ({}/{})".format(current, total) if show_balance else ""
    skip = "\n" if percentage == 100 else ""

    sys.stdout.write(
        "\r{} : |{}{}| {}%{} {}{}".format(prefix, progress, bar, percentage, balance, suffix, skip)
    )
    sys.stdout.flush()


def enumerate_reversed(iterable, start=0, stop=None, step=1):
    """Generator to go through an iterable from back to front with correct indexes without copy of the iterable.

    Source:
        `Stack Overflow <https://stackoverflow.com/a/52195368>`_

    Parameters
    ----------
    iterable : Iterable[_T]
    start, stop, step : int, optional
        Starting and ending index with step size for the iterable

    Yields
    ------
    index, value
        index always refers to the true index of the iterable, independent of start or stop!
    """
    if stop is None:
        stop = len(iterable)

    for index, value in enumerate(reversed(iterable[start:stop:step])):
        index = stop - 1 - index
        yield index, value



""" Wrapper for plots with 2 y axis. Shows only "left plot" in pyqtgraph and merged in tikz

Author:
    Mario Krattenmacher | Mario.Krattenmacher@semimod.de
    Markus Mueller | Markus.Mueller3@tu-dresden.de
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import os
import numpy as np
from pathlib import Path
from DMT.external import build_tex, build_svg, clean_tex_files, build_png, slugify, tex_to_text

try:
    import matplotlib
    import matplotlib.pyplot as plt
    import matplotlib._pylab_helpers

    # rc params spec
    packages = [
        "\\usepackage{amsmath}\n",
        "\\usepackage{mathtools}\n",
        "\\usepackage{amssymb}\n",
        "\\usepackage{siunitx}\n",
        "\\sisetup{range-units=repeat, list-units=repeat, binary-units, exponent-product = \\cdot, print-unity-mantissa=false}\n",
        "\\DeclareSIUnit\\sq{\\ensuremath{\\Box}}\n",
        "\\DeclareSIUnit\\degC{\\degreeCelsius}\n",
    ]
    packages_to_add = []
    str_user_packages = "".join(matplotlib.rcParams["text.latex.preamble"])
    for package in packages:
        if package not in str_user_packages:
            packages_to_add.append(package)

    try:
        matplotlib.rcParams["text.latex.preamble"] += packages_to_add
    except TypeError:
        # must be str ?!?
        matplotlib.rcParams["text.latex.preamble"] += "".join(packages_to_add)

except ImportError:
    print("DMT->Plot: Failed to import plotting module matplotlib.")


class Plot2YAxis(object):
    """Provides a plot with 2 y-Axes

    WARNING: Legend does not work with cyclers :(. You have to set the styles manually.

    """

    def __init__(self, name, plot_left, plot_right, legend_location="upper right"):
        self.name = name

        # do not check anything here!

        self.plot_left = plot_left
        self.plot_right = plot_right

        self.legend_location = legend_location
        self.legend_frame = True

        self.fig = None
        self.ax_left = None
        self.ax_right = None

        self.lines_left = []
        self.lines_right = []

    def plot_py(
        self,
        show=True,
        font_size=None,
        allowGrid=False,
        tight_layout=True,
        figure_size=None,
        use_tex=True,
    ):
        """Plots the 2 axis plot with the same arguments as the usual plot class"""
        matplotlib.rcParams["text.usetex"] = use_tex

        # get a new figure
        self.fig = plt.figure(num=self.name, figsize=figure_size)
        print("init 2 axis figure with name " + self.name + r"\n")
        # setting the window title using the matplotlib figure manager
        # pylint: disable = protected-access
        fig_manager = matplotlib._pylab_helpers.Gcf.get_fig_manager(self.fig.number)
        fig_manager.set_window_title(self.name)  # type: ignore

        if font_size is not None:
            matplotlib.rcParams.update({"font.size": font_size})

        ####################################### left axis
        self.ax_left = self.fig.add_subplot(111)
        # set the line cycler
        self.ax_left.set_prop_cycle(
            self.plot_left._cycler
        )  # does not work properly in combination with legends

        # plotting of the data left
        for _i, dict_line in enumerate(self.plot_left.data):
            # x and y should be numpy vectors, stable conversion is given (more tests necessary)
            x = dict_line["x"]
            y = dict_line["y"]
            if not isinstance(x, np.ndarray):
                try:
                    x = [float(x_a) for x_a in x]
                except TypeError:
                    pass
                x = np.asanyarray(x)
            if not isinstance(y, np.ndarray):
                try:
                    y = [float(y_a) for y_a in y]
                except TypeError:
                    pass
                y = np.asanyarray(y)

            if self.plot_left.x_axis_scale == "log":
                x = np.abs(x)
            if self.plot_left.y_axis_scale == "log":
                y = np.abs(y)

            if use_tex:
                label = dict_line["label"]
            else:
                label = tex_to_text(dict_line["label"])

            if "style" in dict_line and dict_line["style"] is not None:
                try:
                    if (
                        "o" in dict_line["style"]
                    ):  # o is an empty circle from now on! Use '.' for filled points
                        (line,) = self.ax_left.plot(
                            x * self.plot_left.x_scale,
                            y * self.plot_left.y_scale,
                            dict_line["style"],
                            fillstyle="none",
                            label=label,
                            **dict_line["kwargs"],
                        )
                    else:
                        (line,) = self.ax_left.plot(
                            x * self.plot_left.x_scale,
                            y * self.plot_left.y_scale,
                            dict_line["style"],
                            label=label,
                            **dict_line["kwargs"],
                        )
                except ValueError as err:
                    raise ValueError(
                        "Too many values to unpack in plot "
                        + self.plot_left.name
                        + " for line with label "
                        + str(label)
                    ) from err
            else:
                try:
                    (line,) = self.ax_left.plot(
                        x * self.plot_left.x_scale,
                        y * self.plot_left.y_scale,
                        label=label,
                        **dict_line["kwargs"],
                    )
                except ValueError as err:
                    raise ValueError(
                        "Too many values to unpack in plot "
                        + self.plot_left.name
                        + " for line with label "
                        + str(label)
                    ) from err
            self.lines_left.append(line)

        # labels
        if use_tex:
            self.ax_left.set_xlabel(self.plot_left.x_label)
            self.ax_left.set_ylabel(self.plot_left.y_label)
        else:
            self.ax_left.set_xlabel(tex_to_text(self.plot_left.x_label))
            self.ax_left.set_ylabel(tex_to_text(self.plot_left.y_label))

        # set scale and limits
        self.ax_left.set_xscale(self.plot_left.x_axis_scale)
        self.ax_left.set_yscale(self.plot_left.y_axis_scale)

        self.ax_left.set_xlim(self.plot_left.x_limits)
        if not all(lim is None for lim in self.plot_left.x_limits):
            self.ax_left.set_xlim(left=self.plot_left.x_limits[0], right=self.plot_left.x_limits[1])

        if not all(lim is None for lim in self.plot_left.y_limits):
            self.ax_left.set_ylim(bottom=self.plot_left.y_limits[0], top=self.plot_left.y_limits[1])

        ####################################### right axis
        self.ax_right = self.ax_left.twinx()
        # set the line cycler
        self.ax_right.set_prop_cycle(
            self.plot_right._cycler
        )  # does not work properly in combination with legends

        # plotting of the data right
        for _i, dict_line in enumerate(self.plot_right.data):
            # x and y should be numpy vectors, stable conversion is given (more tests necessary)
            x = dict_line["x"]
            y = dict_line["y"]

            if self.plot_right.x_axis_scale == "log":
                x = np.abs(x)
            if self.plot_right.y_axis_scale == "log":
                y = np.abs(y)

            if use_tex:
                label = dict_line["label"]
            else:
                label = tex_to_text(dict_line["label"])
            x_left = self.ax_left.get_xlim()[0] - 1
            y_left = self.ax_left.get_ylim()[0] - 1

            # plot now on right and left axis at the sime time to ensure legend
            if "style" in dict_line and dict_line["style"] is not None:
                try:
                    if (
                        "o" in dict_line["style"]
                    ):  # o is an empty circle from now on! Use '.' for filled points
                        (line,) = self.ax_right.plot(
                            x * self.plot_right.x_scale,
                            y * self.plot_right.y_scale,
                            dict_line["style"],
                            fillstyle="none",
                            label=label,
                            **dict_line["kwargs"],
                        )
                        (_line,) = self.ax_left.plot(
                            x_left,
                            y_left,
                            dict_line["style"],
                            fillstyle="none",
                            label=label,
                            **dict_line["kwargs"],
                        )
                    else:
                        (line,) = self.ax_right.plot(
                            x * self.plot_right.x_scale,
                            y * self.plot_right.y_scale,
                            dict_line["style"],
                            label=label,
                            **dict_line["kwargs"],
                        )
                        (_line,) = self.ax_left.plot(
                            x_left, y_left, dict_line["style"], label=label, **dict_line["kwargs"]
                        )
                except ValueError as err:
                    raise ValueError(
                        "Too many values to unpack in plot "
                        + self.plot_right.name
                        + " for line with label "
                        + str(label)
                    ) from err
            else:
                try:
                    (line,) = self.ax_right.plot(
                        x * self.plot_right.x_scale,
                        y * self.plot_right.y_scale,
                        label=label,
                        **dict_line["kwargs"],
                    )
                    (_line,) = self.ax_left.plot(x_left, y_left, label=label, **dict_line["kwargs"])
                except ValueError as err:
                    raise ValueError(
                        "Too many values to unpack in plot "
                        + self.plot_right.name
                        + " for line with label "
                        + str(label)
                    ) from err
            self.lines_right.append(line)

        # labels (only y label needed)
        if use_tex:
            self.ax_right.set_ylabel(self.plot_right.y_label)
        else:
            self.ax_right.set_ylabel(tex_to_text(self.plot_right.y_label))

        # set scale and limits (only y label needed)
        self.ax_right.set_yscale(self.plot_right.y_axis_scale)

        if not all(lim is None for lim in self.plot_right.y_limits):
            self.ax_right.set_ylim(
                bottom=self.plot_right.y_limits[0], top=self.plot_right.y_limits[1]
            )

        # legend ( only once ! )
        if self.legend_location in [
            "upper right outer",
            "right mid",
        ]:  # not supported in matplotlib
            self.ax_left.legend(loc="upper right", frameon=self.legend_frame)
        else:
            self.ax_left.legend(loc=self.legend_location, frameon=self.legend_frame)

        if allowGrid:
            # Don't allow the axis to be on top of your data
            self.ax_left.set_axisbelow(True)
            # self.ax_right.set_axisbelow(True)

            # Turn on the minor TICKS, which are required for the minor GRID
            self.ax_left.minorticks_on()
            # self.ax_right.minorticks_on()

            # Customize the major grid and show only one grid!
            self.ax_left.grid(which="major", linewidth=0.65, linestyle="-", color=".85")
            # Customize the minor grid and show only one grid!
            self.ax_left.grid(which="minor", linewidth=0.65, linestyle="-", color=".85")

        if tight_layout:
            plt.tight_layout()

        if show:
            plt.show()

    def plot_pyqtgraph(self, *args, **kwargs):
        """At the moment only a pass through to the left plot..."""
        self.plot_left.plot_pyqtgraph(*args, **kwargs)

    def show_pyqtgraph(self, *args, **kwargs):
        self.plot_left.show_pyqtgraph(*args, **kwargs)

    def save_tikz(
        self,
        directory,
        file_name=None,
        width="\\textwidth",
        fontsize="normalsize",
        height=None,
        mark_repeat=1,
        extension=None,
        restrict_left=True,
        restrict_right=True,
        hide_second_ticks=False,
        hide_second_axis=False,
        standalone=False,
        build=False,
        clean=False,
        svg=False,
        png=False,
    ):
        """Save plot in directory and return name of the tikz file. The name of the tikz file will be the figure attribute self.name if not given.

        Parameters
        ----------
        hide_second_ticks : Bool, False
            If true, the y-tick lines of the second plot are not drawn.
        """
        if not isinstance(directory, Path):
            directory = Path(directory)
        os.makedirs(directory, exist_ok=True)

        ext_file = ".tex"
        if extension is not None:
            ext_file = "." + extension

        if file_name is None:
            file_name = slugify(self.name) + ext_file
        elif not file_name.endswith(ext_file):
            file_name = file_name + ext_file

        # create tikz files from the subplots
        # change names as these plots will only exists temporarily
        name_old_left = self.plot_left.name
        num_old_left = self.plot_left.num
        name_old_right = self.plot_right.name
        num_old_right = self.plot_right.num

        self.plot_left.name = name_old_left + "_tmp"
        self.plot_left.num = num_old_left + "_tmp"
        self.plot_right.name = name_old_right + "_tmp"
        self.plot_right.num = num_old_right + "_tmp"

        file_tikz_left = self.plot_left.save_tikz(
            directory,
            width=width,
            height=height,
            mark_repeat=mark_repeat,
            standalone=standalone,
            restrict=restrict_left,
            extension=extension,
        )
        file_tikz_right = self.plot_right.save_tikz(
            directory,
            width=width,
            height=height,
            mark_repeat=mark_repeat,
            standalone=standalone,
            restrict=restrict_right,
            extension=extension,
        )

        # open, read and delete the tikz files
        path_file_left = directory / file_tikz_left
        str_tikz_left = path_file_left.read_text()
        path_file_left.unlink()

        path_file_right = directory / file_tikz_right
        str_tikz_right = path_file_right.read_text()
        path_file_right.unlink()

        self.plot_left.name = name_old_left
        self.plot_left.num = num_old_left
        self.plot_right.name = name_old_right
        self.plot_right.num = num_old_right

        # find color definition from second file

        # add "scale only axis", which makes the y-labels appear correctly independent of figure size
        # add yticks of first y-axis only at left axis, else it looks ugly.
        str_tikz_right = str_tikz_right.replace(r"ytick pos=left", r"ytick pos=right")

        # change str_tikz_right:
        lines_str_tikz_right = str_tikz_right.splitlines()

        # find color definitions from right plot
        color_defs = [line for line in lines_str_tikz_right if "definecolor" in line]

        # find legend from left plot and replace it with labels, these are added later
        legend_entries_left = [
            line
            for line in str_tikz_left.splitlines()
            if "addlegendentry" in line and not "%" in line
        ]
        for i, entry in enumerate(legend_entries_left):
            str_tikz_left = str_tikz_left.replace(
                entry, "\\label{" + file_name.replace(".tex", "_") + str(i) + "}"
            )

        # find start:
        i_start = next(
            i_line for i_line, line in enumerate(lines_str_tikz_right) if r"\begin{axis}[" in line
        )
        lines_to_del = []
        i_line = 0

        for i_line, line in enumerate(lines_str_tikz_right[i_start:]):  # iterate over a copy..
            if line == "]":
                break

            if (
                line.startswith("xlabel")
                or line.startswith("xmajorgrids")
                or line.startswith("xminorgrids")
                or line.startswith("xtick")
                or line.startswith("minor xtick")
                or line.startswith("x ")
            ):
                lines_to_del.append(i_start + i_line)

        str_hide_second_axis = "hide y axis=true,\n"
        if not hide_second_axis:
            str_hide_second_axis = "% " + str_hide_second_axis

        str_hide_second_ticks = ",\n"
        y_label_def = "ylabel style={at={(1.32, 0.5)}, anchor=north},\n"
        if hide_second_ticks:
            str_hide_second_ticks = "ytick=\\empty,\n"
            y_label_def = "ylabel style={at={(1.02, 0.5)}, anchor=north},\n"
        y_label_def = ""

        lines_str_tikz_right.insert(
            i_start + i_line,
            (
                "hide x axis=true,\n"
                + str_hide_second_axis
                + "every outer y axis line/.append style={black},\n"
                + "scale only axis,"
                + "every y tick label/.append style={font=\\color{black}},\n"
                + "every y tick/.append style={black},\n"
                + y_label_def
                + "yticklabel style={text width=0.6em,align=left},\n"
                + "yticklabel pos=right,\n"
                + "axis x line*=bottom,\n"
                + str_hide_second_ticks
                + "axis y line*=right,"
            ),
        )

        for i_del in lines_to_del[::-1]:
            del lines_str_tikz_right[i_del]

        # after axis definition in right plot, add legend entries from left plot
        line_legend_left = "]\n" + "\n".join(
            [
                r"\addlegendimage{/pgfplots/refstyle="
                + file_name.replace(".tex", "_")
                + str(i)
                + "}"
                + entry
                for i, entry in enumerate(legend_entries_left)
            ]
        )
        for i, line in enumerate(lines_str_tikz_right):
            if line == "]":
                lines_str_tikz_right[i] = line_legend_left

        # create one big string from array of strings
        str_tikz_right = "\n".join(lines_str_tikz_right)

        # merge the strings
        # color defs from right plot
        color_defs = "\n".join(color_defs) + "\n\\begin{axis}"
        str_tikz_left = str_tikz_left.replace(r"\begin{axis}", color_defs)

        # from left plot everything including \end{axis} and before end{tikzpicture}
        i_left_end = str_tikz_left.find(r"\end{tikzpicture}")
        str_tikz_picture = str_tikz_left[:i_left_end]

        # append right plot starting after \begin{tikzpicture} and before \begin{axis}
        i_right_start = str_tikz_right.find(r"\begin{axis}")
        str_tikz_picture += "\n" + str_tikz_right[i_right_start:]

        # if legend location is upper right outer, shift to right
        if hide_second_ticks:
            str_tikz_picture = str_tikz_picture.replace(
                "at={(1.02,1.00)}, anchor=north west,", "at={(1.07,1.00)}, anchor=north west,"
            )
        else:
            str_tikz_picture = str_tikz_picture.replace(
                "at={(1.02,1.00)}, anchor=north west,", "at={(1.12,1.00)}, anchor=north west,"
            )

        # having y-major ticks makes these plots ugly. Either rescale the ticks or just remove them. At the moment, just remove
        str_tikz_picture = str_tikz_picture.replace("ymajorgrids", "")

        # correct sizes
        if width is not None:
            str_tikz_picture = str_tikz_picture.replace(
                "\\setlength\\figurewidth{60mm}\n", "\\setlength\\figurewidth{" + width + "}\n"
            )
        if height is not None:
            str_tikz_picture = str_tikz_picture.replace(
                "\\setlength\\figureheight{60mm}\n", "\\setlength\\figureheight{" + height + "}\n"
            )

        path_file = directory / file_name
        path_file.write_text(str_tikz_picture)

        if build:
            if not standalone:
                raise IOError(
                    "To build a single TikZ picture, you should also activate standalone!"
                )

            if svg:
                build_svg(path_file, wait=clean)
                ending_to_keep = ".svg"
            elif png:
                build_tex(path_file, wait=clean)
                build_png(path_file, wait=clean)
                ending_to_keep = ".png"
            else:
                build_tex(path_file, wait=clean)
                ending_to_keep = ".pdf"

            if clean:
                clean_tex_files(
                    directory, file_name.replace(ext_file, ""), keep=(ending_to_keep, ext_file)
                )

        return file_name
